# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
from typing import Any

from .task import Task
from ..work import Work


class TaskHandler(Work[Task]):
    """Task handler."""

    @staticmethod
    def create(*args: Any) -> Task:
        """Work core doesn't know how to create work objects for us.
        Example, for task module scenario, it doesn't know how to create
        Task objects for us."""
        return Task(*args)



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import io
import os
import inspect
import logging
import importlib
import itertools
# pylint: disable=ungrouped-imports
import importlib.util
from types import ModuleType
from typing import Any, Dict, List, Tuple, Union, Optional

from .utils import text_, bytes_
from .constants import DOT, COMMA, DEFAULT_ABC_PLUGINS


logger = logging.getLogger(__name__)


class Plugins:
    """Common utilities for plugin discovery."""

    @staticmethod
    def resolve_plugin_flag(flag_plugins: Any, opt_plugins: Optional[Any] = None) -> List[Union[bytes, type]]:
        if isinstance(flag_plugins, list):
            requested_plugins = list(
                itertools.chain.from_iterable([
                    p.split(text_(COMMA)) for p in list(
                        itertools.chain.from_iterable(flag_plugins),
                    )
                ]),
            )
        else:
            requested_plugins = flag_plugins.split(text_(COMMA))
        return [
            p if isinstance(p, type) else bytes_(p)
            for p in (opt_plugins if opt_plugins is not None else requested_plugins)
            if not (isinstance(p, str) and len(p) == 0)
        ]

    @staticmethod
    def discover(input_args: List[str]) -> None:
        """Search for external plugin found in command line arguments,
        then iterates over each value and discover/import the plugin.
        """
        for i, f in enumerate(input_args):
            if f in ('--plugin', '--plugins', '--auth-plugin'):
                v = input_args[i + 1]
                parts = v.split(',')
                for part in parts:
                    Plugins.importer(bytes_(part))

    @staticmethod
    def load(
        plugins: List[Union[bytes, type]],
        abc_plugins: Optional[List[str]] = None,
    ) -> Dict[bytes, List[type]]:
        """Accepts a list Python modules, scans them to identify
        if they are an implementation of abstract plugin classes and
        returns a dictionary of matching plugins for each abstract class.
        """
        p: Dict[bytes, List[type]] = {}
        for abc_plugin in (abc_plugins or DEFAULT_ABC_PLUGINS):
            p[bytes_(abc_plugin)] = []
        for plugin_ in plugins:
            klass, module_name = Plugins.importer(plugin_)
            assert klass and module_name
            mro = list(inspect.getmro(klass))
            # Find the base plugin class that
            # this plugin_ is implementing
            base_klass = None
            for k in mro:
                if bytes_(k.__qualname__) in p:
                    base_klass = k
                    break
            if base_klass is None:
                raise ValueError('%s is NOT a valid plugin' % text_(plugin_))
            if klass not in p[bytes_(base_klass.__qualname__)]:
                p[bytes_(base_klass.__qualname__)].append(klass)
            logger.info('Loaded plugin %s.%s', module_name, klass.__qualname__)
        # print(p)
        return p

    @staticmethod
    def importer(plugin: Union[bytes, type]) -> Tuple[type, str]:
        """Import and returns the plugin."""
        if isinstance(plugin, type):
            if inspect.isclass(plugin):
                return (plugin, plugin.__module__ or '__main__')
            raise ValueError('%s is not a valid reference to a plugin class' % text_(plugin))
        plugin_ = text_(plugin.strip())
        assert plugin_ != ''
        path = plugin_.split(text_(DOT))
        klass = None

        def locate_klass(klass_module_name: str, klass_path: List[str]) -> Union[type, None]:
            klass_module_name = klass_module_name.replace(os.path.sep, text_(DOT))
            try:
                klass_module = importlib.import_module(klass_module_name)
            except ModuleNotFoundError:
                return None
            klass_container: Union[ModuleType, type] = klass_module
            for klass_path_part in klass_path:
                try:
                    klass_container = getattr(klass_container, klass_path_part)
                except AttributeError:
                    return None
            if not isinstance(klass_container, type) or not inspect.isclass(klass_container):
                return None
            return klass_container

        module_name = None
        for module_name_parts in range(len(path) - 1, 0, -1):
            module_name = '.'.join(path[0:module_name_parts])
            klass = locate_klass(module_name, path[module_name_parts:])
            if klass:
                break
        if klass is None:
            module_name = '__main__'
            klass = locate_klass(module_name, path)
        if klass is None or module_name is None:
            raise ValueError('%s is not resolvable as a plugin class' % text_(plugin))
        return (klass, module_name)

    @staticmethod
    def from_bytes(pyc: bytes, name: str) -> ModuleType:
        code_stream = io.BytesIO(pyc)
        spec = importlib.util.spec_from_loader(
            name,
            loader=None,
            origin='dynamic',
            is_package=True,
        )
        assert spec is not None
        mod = importlib.util.module_from_spec(spec)
        code_stream.seek(0)
        # pylint: disable=exec-used
        exec(code_stream.read(), mod.__dict__)  # noqa: S102
        return mod



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import asyncio
from typing import Any, Optional
from multiprocessing import connection
from multiprocessing.reduction import recv_handle

from .fd import ThreadlessFdExecutor


class RemoteFdExecutor(ThreadlessFdExecutor[connection.Connection]):
    """A threadless executor implementation which receives work over a connection.

    NOTE: RemoteExecutor uses ``recv_handle`` to accept file descriptors.

    TODO: Refactor and abstract ``recv_handle`` part so that a threaded
    remote executor can also accept work over a connection.  Currently,
    remote executors must be running in a process.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._loop: Optional[asyncio.AbstractEventLoop] = None

    @property
    def loop(self) -> Optional[asyncio.AbstractEventLoop]:
        if self._loop is None:
            self._loop = asyncio.get_event_loop_policy().get_event_loop()
        return self._loop

    def receive_from_work_queue(self) -> bool:
        # Acceptor will not send address for
        # unix socket domain environments.
        addr = None
        if not self.flags.unix_socket_path:
            addr = self.work_queue.recv()
        fileno = recv_handle(self.work_queue)
        self.work(fileno, addr, None)
        return False

    def work_queue_fileno(self) -> Optional[int]:
        return self.work_queue.fileno()

    def close_work_queue(self) -> None:
        self.work_queue.close()



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.

    .. spelling::

       http
       Submodules
"""
from .chunk import ChunkParser, chunkParserStates
from .types import httpParserTypes, httpParserStates
from .parser import HttpParser
from .protocol import PROXY_PROTOCOL_V2_SIGNATURE, ProxyProtocol


__all__ = [
    'HttpParser',
    'httpParserTypes',
    'httpParserStates',
    'ChunkParser',
    'chunkParserStates',
    'ProxyProtocol',
    'PROXY_PROTOCOL_V2_SIGNATURE',
]



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
from typing import List, Tuple, Optional, NamedTuple

from ...common.utils import bytes_, find_http_line
from ...common.constants import CRLF, DEFAULT_BUFFER_SIZE


ChunkParserStates = NamedTuple(
    'ChunkParserStates', [
        ('WAITING_FOR_SIZE', int),
        ('WAITING_FOR_DATA', int),
        ('COMPLETE', int),
    ],
)
chunkParserStates = ChunkParserStates(1, 2, 3)


class ChunkParser:
    """HTTP chunked encoding response parser."""

    def __init__(self) -> None:
        self.state = chunkParserStates.WAITING_FOR_SIZE
        self.body: bytes = b''  # Parsed chunks
        self.chunk: bytes = b''  # Partial chunk received
        # Expected size of next following chunk
        self.size: Optional[int] = None

    def parse(self, raw: memoryview) -> memoryview:
        more = len(raw) > 0
        while more and self.state != chunkParserStates.COMPLETE:
            more, raw = self.process(raw.tobytes())
        return raw

    def process(self, raw: bytes) -> Tuple[bool, memoryview]:
        if self.state == chunkParserStates.WAITING_FOR_SIZE:
            # Consume prior chunk in buffer
            # in case chunk size without CRLF was received
            raw = self.chunk + raw
            self.chunk = b''
            # Extract following chunk data size
            line, raw = find_http_line(raw)
            # CRLF not received or Blank line was received.
            if line is None or line.strip() == b'':
                self.chunk = raw
                raw = b''
            else:
                self.size = int(line, 16)
                self.state = chunkParserStates.WAITING_FOR_DATA
        elif self.state == chunkParserStates.WAITING_FOR_DATA:
            assert self.size is not None
            remaining = self.size - len(self.chunk)
            self.chunk += raw[:remaining]
            raw = raw[remaining:]
            if len(self.chunk) == self.size:
                raw = raw[len(CRLF):]
                self.body += self.chunk
                if self.size == 0:
                    self.state = chunkParserStates.COMPLETE
                else:
                    self.state = chunkParserStates.WAITING_FOR_SIZE
                self.chunk = b''
                self.size = None
        return len(raw) > 0, memoryview(raw)

    @staticmethod
    def to_chunks(raw: bytes, chunk_size: int = DEFAULT_BUFFER_SIZE) -> bytes:
        chunks: List[bytes] = []
        for i in range(0, len(raw), chunk_size):
            chunk = raw[i: i + chunk_size]
            chunks.append(bytes_('{:x}'.format(len(chunk))))
            chunks.append(chunk)
        chunks.append(bytes_('{:x}'.format(0)))
        chunks.append(b'')
        return CRLF.join(chunks) + CRLF



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.

    .. spelling::

       api
"""
import json
from typing import Optional

from ..http.proxy import HttpProxyBasePlugin
from ..http.parser import HttpParser
from ..common.utils import text_, bytes_
from ..http.responses import NOT_FOUND_RESPONSE_PKT, okResponse


class ProposedRestApiPlugin(HttpProxyBasePlugin):
    """Mock responses for your upstream REST API.

    Used to test and develop client side applications
    without need of an actual upstream REST API server.

    Returns proposed REST API mock responses to the client
    without establishing upstream connection.

    Note: This plugin won't work if your client is making
    HTTPS connection to ``api.example.com``.
    """

    API_SERVER = b'api.example.com'

    REST_API_SPEC = {
        b'/v1/users/': {
            'count': 2,
            'next': None,
            'previous': None,
            'results': [
                {
                    'email': 'you@example.com',
                    'groups': [],
                    'url': text_(API_SERVER) + '/v1/users/1/',
                    'username': 'admin',
                },
                {
                    'email': 'someone@example.com',
                    'groups': [],
                    'url': text_(API_SERVER) + '/v1/users/2/',
                    'username': 'someone',
                },
            ],
        },
    }

    def before_upstream_connection(
            self, request: HttpParser,
    ) -> Optional[HttpParser]:
        # Return None to disable establishing connection to upstream
        # Most likely our api.example.com won't even exist under development
        # scenario
        return None

    def handle_client_request(
            self, request: HttpParser,
    ) -> Optional[HttpParser]:
        if request.host != self.API_SERVER:
            return request
        assert request.path
        if request.path in self.REST_API_SPEC:
            self.client.queue(
                okResponse(
                    headers={b'Content-Type': b'application/json'},
                    content=bytes_(
                        json.dumps(
                            self.REST_API_SPEC[request.path],
                        ),
                    ),
                ),
            )
        else:
            self.client.queue(NOT_FOUND_RESPONSE_PKT)
        return None



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import json
import time
from typing import TYPE_CHECKING, Any, Dict

from ..websocket import WebsocketFrame
from ...core.event import eventNames
from ...common.utils import bytes_
from ...common.constants import (
    PROXY_PY_START_TIME, DEFAULT_DEVTOOLS_DOC_URL, DEFAULT_DEVTOOLS_FRAME_ID,
    DEFAULT_DEVTOOLS_LOADER_ID,
)


if TYPE_CHECKING:   # pragma: no cover
    from ..connection import HttpClientConnection


class CoreEventsToDevtoolsProtocol:
    """Open in Chrome

    ``devtools://devtools/bundled/inspector.html?ws=localhost:8899/devtools``
    """

    RESPONSES: Dict[str, bytes] = {}

    @staticmethod
    def transformer(
        client: 'HttpClientConnection',
        event: Dict[str, Any],
    ) -> None:
        event_name = event['event_name']
        if event_name == eventNames.REQUEST_COMPLETE:
            data = CoreEventsToDevtoolsProtocol.request_complete(event)
        elif event_name == eventNames.RESPONSE_HEADERS_COMPLETE:
            data = CoreEventsToDevtoolsProtocol.response_headers_complete(
                event,
            )
        elif event_name == eventNames.RESPONSE_CHUNK_RECEIVED:
            data = CoreEventsToDevtoolsProtocol.response_chunk_received(event)
        elif event_name == eventNames.RESPONSE_COMPLETE:
            data = CoreEventsToDevtoolsProtocol.response_complete(event)
        else:
            # drop core events unrelated to Devtools
            return
        client.queue(
            memoryview(
                WebsocketFrame.text(
                    bytes_(
                        json.dumps(data),
                    ),
                ),
            ),
        )

    @staticmethod
    def request_complete(event: Dict[str, Any]) -> Dict[str, Any]:
        now = time.time()
        return {
            'method': 'Network.requestWillBeSent',
            'params': {
                'requestId': event['request_id'],
                'frameId': DEFAULT_DEVTOOLS_FRAME_ID,
                'loaderId': DEFAULT_DEVTOOLS_LOADER_ID,
                'documentURL': DEFAULT_DEVTOOLS_DOC_URL,
                'timestamp': now - PROXY_PY_START_TIME,
                'wallTime': now,
                'hasUserGesture': False,
                'type': event['event_payload']['headers']['content-type']
                if 'content-type' in event['event_payload']['headers']
                else 'Other',
                'request': {
                    'url': event['event_payload']['url'],
                    'method': event['event_payload']['method'],
                    'headers': event['event_payload']['headers'],
                    'postData': event['event_payload']['body'],
                    'initialPriority': 'High',
                    'urlFragment': '',
                    'mixedContentType': 'none',
                },
                'initiator': {
                    'type': 'other',
                },
            },
        }

    @staticmethod
    def response_headers_complete(event: Dict[str, Any]) -> Dict[str, Any]:
        return {
            'method': 'Network.responseReceived',
            'params': {
                'requestId': event['request_id'],
                'frameId': DEFAULT_DEVTOOLS_FRAME_ID,
                'loaderId': DEFAULT_DEVTOOLS_LOADER_ID,
                'timestamp': time.time(),
                'type': event['event_payload']['headers']['content-type']
                if event['event_payload']['headers'].has_header('content-type')
                else 'Other',
                'response': {
                    'url': '',
                    'status': '',
                    'statusText': '',
                    'headers': '',
                    'headersText': '',
                    'mimeType': '',
                    'connectionReused': True,
                    'connectionId': '',
                    'encodedDataLength': '',
                    'fromDiskCache': False,
                    'fromServiceWorker': False,
                    'timing': {
                        'requestTime': '',
                        'proxyStart': -1,
                        'proxyEnd': -1,
                        'dnsStart': -1,
                        'dnsEnd': -1,
                        'connectStart': -1,
                        'connectEnd': -1,
                        'sslStart': -1,
                        'sslEnd': -1,
                        'workerStart': -1,
                        'workerReady': -1,
                        'sendStart': 0,
                        'sendEnd': 0,
                        'receiveHeadersEnd': 0,
                    },
                    'requestHeaders': '',
                    'remoteIPAddress': '',
                    'remotePort': '',
                },
            },
        }

    @staticmethod
    def response_chunk_received(event: Dict[str, Any]) -> Dict[str, Any]:
        return {
            'method': 'Network.dataReceived',
            'params': {
                'requestId': event['request_id'],
                'timestamp': time.time(),
                'dataLength': event['event_payload']['chunk_size'],
                'encodedDataLength': event['event_payload']['encoded_chunk_size'],
            },
        }

    @staticmethod
    def response_complete(event: Dict[str, Any]) -> Dict[str, Any]:
        return {
            'method': 'Network.loadingFinished',
            'params': {
                'requestId': event['request_id'],
                'timestamp': time.time(),
                'encodedDataLength': event['event_payload']['encoded_response_size'],
                'shouldReportCorbBlocking': False,
            },
        }



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.

    .. spelling::

       Cloudflare
       ws
       onmessage
       httpbin
       localhost
       Lua
"""
from .cache import CacheResponsesPlugin, BaseCacheResponsesPlugin
from .shortlink import ShortLinkPlugin
from .proxy_pool import ProxyPoolPlugin
from .program_name import ProgramNamePlugin
from .mock_rest_api import ProposedRestApiPlugin
from .reverse_proxy import ReverseProxyPlugin
from .cloudflare_dns import CloudflareDnsResolverPlugin
from .modify_post_data import ModifyPostDataPlugin
from .web_server_route import WebServerPlugin
from .man_in_the_middle import ManInTheMiddlePlugin
from .filter_by_upstream import FilterByUpstreamHostPlugin
from .custom_dns_resolver import CustomDnsResolverPlugin
from .filter_by_client_ip import FilterByClientIpPlugin
from .filter_by_url_regex import FilterByURLRegexPlugin
from .modify_chunk_response import ModifyChunkResponsePlugin
from .modify_request_header import ModifyRequestHeaderPlugin
from .redirect_to_custom_server import RedirectToCustomServerPlugin
from .tls_intercept_conditionally import TlsInterceptConditionallyPlugin


__all__ = [
    'CacheResponsesPlugin',
    'BaseCacheResponsesPlugin',
    'FilterByUpstreamHostPlugin',
    'ManInTheMiddlePlugin',
    'ProposedRestApiPlugin',
    'ModifyPostDataPlugin',
    'RedirectToCustomServerPlugin',
    'ShortLinkPlugin',
    'WebServerPlugin',
    'ReverseProxyPlugin',
    'ProxyPoolPlugin',
    'FilterByClientIpPlugin',
    'ModifyChunkResponsePlugin',
    'FilterByURLRegexPlugin',
    'CustomDnsResolverPlugin',
    'CloudflareDnsResolverPlugin',
    'ProgramNamePlugin',
    'ModifyRequestHeaderPlugin',
    'TlsInterceptConditionallyPlugin',
]



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.

    .. spelling::

       devtools
"""
import json
import logging
from typing import Any, Dict, List, Tuple

from ..parser import HttpParser
from ..server import HttpWebServerBasePlugin, httpProtocolTypes
from ..websocket import WebsocketFrame, websocketOpcodes
from .transformer import CoreEventsToDevtoolsProtocol
from ...core.event import EventSubscriber
from ...common.flag import flags
from ...common.utils import text_, bytes_
from ...common.constants import (
    DEFAULT_ENABLE_DEVTOOLS, DEFAULT_DEVTOOLS_DOC_URL,
    DEFAULT_DEVTOOLS_WS_PATH,
)


logger = logging.getLogger(__name__)


flags.add_argument(
    '--enable-devtools',
    action='store_true',
    default=DEFAULT_ENABLE_DEVTOOLS,
    help='Default: False.  Enables integration with Chrome Devtool Frontend. Also see --devtools-ws-path.',
)

flags.add_argument(
    '--devtools-ws-path',
    type=str,
    default=DEFAULT_DEVTOOLS_WS_PATH,
    help='Default: /devtools.  Only applicable '
    'if --enable-devtools is used.',
)


class DevtoolsProtocolPlugin(HttpWebServerBasePlugin):
    """Speaks DevTools protocol with client over websocket.

    - It responds to DevTools client request methods and also
      relay proxy.py core events to the client.
    - Core events are transformed into DevTools protocol format before
      dispatching to client.
    - Core events unrelated to DevTools protocol are dropped.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.subscriber = EventSubscriber(
            self.event_queue,
            callback=lambda event: CoreEventsToDevtoolsProtocol.transformer(
                self.client, event,
            ),
        )

    def routes(self) -> List[Tuple[int, str]]:
        return [
            (httpProtocolTypes.WEBSOCKET, text_(self.flags.devtools_ws_path)),
        ]

    def handle_request(self, request: HttpParser) -> None:
        raise NotImplementedError('This should have never been called')

    def on_websocket_open(self) -> None:
        self.subscriber.setup()

    def on_websocket_message(self, frame: WebsocketFrame) -> None:
        try:
            assert frame.data
            message = json.loads(frame.data)
        except UnicodeDecodeError:
            logger.error(frame.data)
            logger.info(frame.opcode)
            return
        self.handle_devtools_message(message)

    def on_client_connection_close(self) -> None:
        self.subscriber.shutdown()

    def handle_devtools_message(self, message: Dict[str, Any]) -> None:
        frame = WebsocketFrame()
        frame.fin = True
        frame.opcode = websocketOpcodes.TEXT_FRAME
        # logger.info(message)
        method = message['method']
        if method in (
            'Page.canScreencast',
            'Network.canEmulateNetworkConditions',
            'Emulation.canEmulate',
        ):
            data: Dict[str, Any] = {
                'result': False,
            }
        elif method == 'Page.getResourceTree':
            data = {
                'result': {
                    'frameTree': {
                        'frame': {
                            'id': 1,
                            'url': DEFAULT_DEVTOOLS_DOC_URL,
                            'mimeType': 'other',
                        },
                        'childFrames': [],
                        'resources': [],
                    },
                },
            }
        elif method == 'Network.getResponseBody':
            connection_id = message['params']['requestId']
            data = {
                'result': {
                    'body': text_(CoreEventsToDevtoolsProtocol.RESPONSES[connection_id]),
                    'base64Encoded': False,
                },
            }
        else:
            logging.warning('Unhandled devtools method %s', method)
            data = {}
        data['id'] = message['id']
        frame.data = bytes_(json.dumps(data))
        self.client.queue(memoryview(frame.build()))



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.

    .. spelling::

       http
       iterable
"""
from typing import NamedTuple


HttpParserStates = NamedTuple(
    'HttpParserStates', [
        ('INITIALIZED', int),
        ('LINE_RCVD', int),
        ('RCVING_HEADERS', int),
        ('HEADERS_COMPLETE', int),
        ('RCVING_BODY', int),
        ('COMPLETE', int),
    ],
)
httpParserStates = HttpParserStates(1, 2, 3, 4, 5, 6)

HttpParserTypes = NamedTuple(
    'HttpParserTypes', [
        ('REQUEST_PARSER', int),
        ('RESPONSE_PARSER', int),
    ],
)
httpParserTypes = HttpParserTypes(1, 2)



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import os
import glob
from typing import Any, Dict
from pathlib import Path
from multiprocessing.synchronize import Lock

from ...core.event import EventQueue, EventSubscriber, eventNames
from ...common.constants import DEFAULT_METRICS_DIRECTORY_PATH


class MetricsStorage:

    def __init__(self, lock: Lock) -> None:
        self._lock = lock

    def get_counter(self, name: str) -> float:
        with self._lock:
            return self._get_counter(name)

    def _get_counter(self, name: str) -> float:
        path = os.path.join(DEFAULT_METRICS_DIRECTORY_PATH, f'{name}.counter')
        if not os.path.exists(path):
            return 0
        return float(Path(path).read_text(encoding='utf-8').strip())

    def incr_counter(self, name: str, by: float = 1.0) -> None:
        with self._lock:
            self._incr_counter(name, by)

    def _incr_counter(self, name: str, by: float = 1.0) -> None:
        current = self._get_counter(name)
        path = os.path.join(DEFAULT_METRICS_DIRECTORY_PATH, f'{name}.counter')
        Path(path).write_text(str(current + by), encoding='utf-8')

    def get_gauge(self, name: str) -> float:
        with self._lock:
            return self._get_gauge(name)

    def _get_gauge(self, name: str) -> float:
        path = os.path.join(DEFAULT_METRICS_DIRECTORY_PATH, f'{name}.gauge')
        if not os.path.exists(path):
            return 0
        return float(Path(path).read_text(encoding='utf-8').strip())

    def set_gauge(self, name: str, value: float) -> None:
        """Stores a single values."""
        with self._lock:
            self._set_gauge(name, value)

    def _set_gauge(self, name: str, value: float) -> None:
        path = os.path.join(DEFAULT_METRICS_DIRECTORY_PATH, f'{name}.gauge')
        with open(path, 'w', encoding='utf-8') as g:
            g.write(str(value))


class MetricsEventSubscriber:

    def __init__(self, event_queue: EventQueue, metrics_lock: Lock) -> None:
        """Aggregates metric events pushed by proxy.py core and plugins.

        1) Metrics are stored and managed by multiprocessing safe MetricsStorage
        2) Collection must be done via MetricsWebServerPlugin endpoint
        """
        self.storage = MetricsStorage(metrics_lock)
        self.subscriber = EventSubscriber(
            event_queue,
            callback=lambda event: MetricsEventSubscriber.callback(self.storage, event),
        )

    def setup(self) -> None:
        self._setup_metrics_directory()
        self.subscriber.setup()

    def shutdown(self) -> None:
        self.subscriber.shutdown()

    def __enter__(self) -> 'MetricsEventSubscriber':
        self.setup()
        return self

    def __exit__(self, *args: Any) -> None:
        self.shutdown()

    @staticmethod
    def callback(storage: MetricsStorage, event: Dict[str, Any]) -> None:
        if event['event_name'] == eventNames.WORK_STARTED:
            storage.incr_counter('work_started')
        elif event['event_name'] == eventNames.REQUEST_COMPLETE:
            storage.incr_counter('request_complete')
        elif event['event_name'] == eventNames.WORK_FINISHED:
            storage.incr_counter('work_finished')
        else:
            print('Unhandled', event)

    def _setup_metrics_directory(self) -> None:
        os.makedirs(DEFAULT_METRICS_DIRECTORY_PATH, exist_ok=True)
        patterns = ['*.counter', '*.gauge']
        for pattern in patterns:
            files = glob.glob(os.path.join(DEFAULT_METRICS_DIRECTORY_PATH, pattern))
            for file_path in files:
                try:
                    os.remove(file_path)
                except OSError as e:
                    print(f'Error deleting file {file_path}: {e}')



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.

    .. spelling::

       http
       iterable
"""
from typing import NamedTuple


HttpProtocols = NamedTuple(
    'HttpProtocols', [
        ('UNKNOWN', int),
        # Web server handling HTTP/1.0, HTTP/1.1, HTTP/2, HTTP/3
        # over plain Text or encrypted connection with clients
        ('WEB_SERVER', int),
        # Proxies handling HTTP/1.0, HTTP/1.1, HTTP/2 protocols
        # over plain text connection or encrypted connection
        # with clients
        ('HTTP_PROXY', int),
        # Proxies handling SOCKS4, SOCKS4a, SOCKS5 protocol
        ('SOCKS_PROXY', int),
    ],
)

httpProtocols = HttpProtocols(1, 2, 3, 4)



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
from typing import NamedTuple


# Name of the events that eventing framework supports.
#
# Ideally this must be configurable via command line or
# at-least extendable via plugins.
EventNames = NamedTuple(
    'EventNames', [
        ('SUBSCRIBE', int),
        ('SUBSCRIBED', int),
        ('UNSUBSCRIBE', int),
        ('UNSUBSCRIBED', int),
        ('DISPATCHER_SHUTDOWN', int),
        ('WORK_STARTED', int),
        ('WORK_FINISHED', int),
        ('REQUEST_COMPLETE', int),
        ('RESPONSE_HEADERS_COMPLETE', int),
        ('RESPONSE_CHUNK_RECEIVED', int),
        ('RESPONSE_COMPLETE', int),
    ],
)
eventNames = EventNames(1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11)



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
from .task import Task
from .local import LocalTaskExecutor, ThreadedTaskExecutor
from .remote import RemoteTaskExecutor, SingleProcessTaskExecutor
from .handler import TaskHandler


__all__ = [
    'Task',
    'TaskHandler',
    'LocalTaskExecutor',
    'ThreadedTaskExecutor',
    'RemoteTaskExecutor',
    'SingleProcessTaskExecutor',
]



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import os
import struct
import logging
from typing import Tuple, Optional

from .pretty import pretty_hexlify


logger = logging.getLogger(__name__)


class TlsHelloRequest:
    """TLS Hello Request"""

    def __init__(self) -> None:
        self.data: Optional[bytes] = None

    def parse(self, raw: bytes) -> None:
        self.data = raw

    def build(self) -> bytes:
        assert self.data
        return self.data


class TlsClientHello:
    """TLS Client Hello"""

    def __init__(self) -> None:
        self.protocol_version: Optional[bytes] = None
        self.random: Optional[bytes] = None
        self.session_id: Optional[bytes] = None
        self.cipher_suite: Optional[bytes] = None
        self.compression_method: Optional[bytes] = None
        self.extension: Optional[bytes] = None

    def parse(self, raw: bytes) -> Tuple[bool, bytes]:
        try:
            idx = 0
            length = len(raw)
            self.protocol_version = raw[idx:idx + 2]
            idx += 2
            self.random = raw[idx:idx + 32]
            idx += 32
            session_length = raw[idx]
            self.session_id = raw[idx: idx + 1 + session_length]
            idx += 1 + session_length
            cipher_suite_length, = struct.unpack('!H', raw[idx: idx + 2])
            self.cipher_suite = raw[idx: idx + 2 + cipher_suite_length]
            idx += 2 + cipher_suite_length
            compression_method_length = raw[idx]
            self.compression_method = raw[
                idx: idx +
                1 + compression_method_length
            ]
            idx += 1 + compression_method_length
            # extension
            if idx == length:
                self.extension = b''
            else:
                extension_length, = struct.unpack('!H', raw[idx: idx + 2])
                self.extension = raw[idx: idx + 2 + extension_length]
                idx += 2 + extension_length
            return True, raw[idx:]
        except Exception as e:
            logger.exception(e)
            return False, raw

    def build(self) -> bytes:
        # calculate length
        return b''.join([
            bs for bs in (
                self.protocol_version, self.random, self.session_id, self.cipher_suite,
                self.compression_method, self.extension,
            ) if bs is not None
        ])

    def format(self) -> str:
        parts = []
        parts.append(
            'Protocol Version: %s' % (
                pretty_hexlify(self.protocol_version)
                if self.protocol_version is not None
                else ''
            ),
        )
        parts.append(
            'Random: %s' % (
                pretty_hexlify(self.random)
                if self.random is not None else ''
            ),
        )
        parts.append(
            'Session ID: %s' % (
                pretty_hexlify(self.session_id)
                if self.session_id is not None
                else ''
            ),
        )
        parts.append(
            'Cipher Suite: %s' % (
                pretty_hexlify(self.cipher_suite)
                if self.cipher_suite is not None
                else ''
            ),
        )
        parts.append(
            'Compression Method: %s' % (
                pretty_hexlify(self.compression_method)
                if self.compression_method is not None
                else ''
            ),
        )
        parts.append(
            'Extension: %s' % (
                pretty_hexlify(self.extension)
                if self.extension is not None
                else ''
            ),
        )
        return os.linesep.join(parts)


class TlsServerHello:
    """TLS Server Hello"""

    def __init__(self) -> None:
        self.protocol_version: Optional[bytes] = None
        self.random: Optional[bytes] = None
        self.session_id: Optional[bytes] = None
        self.cipher_suite: Optional[bytes] = None
        self.compression_method: Optional[bytes] = None
        self.extension: Optional[bytes] = None

    def parse(self, raw: bytes) -> Tuple[bool, bytes]:
        try:
            idx = 0
            length = len(raw)
            self.protocol_version = raw[idx:idx + 2]
            idx += 2
            self.random = raw[idx:idx + 32]
            idx += 32
            session_length = raw[idx]
            self.session_id = raw[idx: idx + 1 + session_length]
            idx += 1 + session_length
            self.cipher_suite = raw[idx: idx + 2]
            idx += 2
            compression_method_length = raw[idx]
            self.compression_method = raw[
                idx: idx +
                1 + compression_method_length
            ]
            idx += 1 + compression_method_length
            # extension
            if idx == length:
                self.extension = b''
            else:
                extension_length, = struct.unpack('!H', raw[idx: idx + 2])
                self.extension = raw[idx: idx + 2 + extension_length]
                idx += 2 + extension_length
            return True, raw[idx:]
        except Exception as e:
            logger.exception(e)
            return False, raw

    def build(self) -> bytes:
        return b''.join([
            bs for bs in (
                self.protocol_version, self.random, self.session_id, self.cipher_suite,
                self.compression_method, self.extension,
            ) if bs is not None
        ])

    def format(self) -> str:
        parts = []
        parts.append(
            'Protocol Version: %s' % (
                pretty_hexlify(self.protocol_version)
                if self.protocol_version is not None
                else ''
            ),
        )
        parts.append(
            'Random: %s' % (
                pretty_hexlify(self.random)
                if self.random is not None
                else ''
            ),
        )
        parts.append(
            'Session ID: %s' % (
                pretty_hexlify(self.session_id)
                if self.session_id is not None
                else ''
            ),
        )
        parts.append(
            'Cipher Suite: %s' % (
                pretty_hexlify(self.cipher_suite)
                if self.cipher_suite is not None
                else ''
            ),
        )
        parts.append(
            'Compression Method: %s' % (
                pretty_hexlify(self.compression_method)
                if self.compression_method is not None
                else ''
            ),
        )
        parts.append(
            'Extension: %s' % (
                pretty_hexlify(self.extension)
                if self.extension is not None
                else ''
            ),
        )
        return os.linesep.join(parts)


class TlsServerHelloDone:
    """TLS Server Hello Done"""

    def __init__(self) -> None:
        self.data: Optional[bytes] = None

    def parse(self, raw: bytes) -> Tuple[bool, bytes]:
        return False, raw

    def build(self) -> bytes:
        assert self.data
        return self.data



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""


class Task:
    """Task object which known how to process the payload."""

    def __init__(self, payload: bytes) -> None:
        self.payload = payload
        print(payload)



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.

    .. spelling::

       pre
"""
import socket
import logging
import argparse
import selectors
import threading
import multiprocessing
import multiprocessing.synchronize
from typing import Dict, List, Tuple, Optional
from multiprocessing import connection
from multiprocessing.reduction import recv_handle

from ..work import start_threaded_work, delegate_work_to_pool
from ..event import EventQueue
from ..work.fd import LocalFdExecutor
from ...common.flag import flags
from ...common.types import HostPort
from ...common.logger import Logger
from ...common.backports import NonBlockingQueue
from ...common.constants import DEFAULT_LOCAL_EXECUTOR


logger = logging.getLogger(__name__)


flags.add_argument(
    '--local-executor',
    type=int,
    default=int(DEFAULT_LOCAL_EXECUTOR),
    help='Default: ' + ('1' if DEFAULT_LOCAL_EXECUTOR else '0') + '.  ' +
    'Enabled by default.  Use 0 to disable.  When enabled acceptors ' +
    'will make use of local (same process) executor instead of distributing load across ' +
    'remote (other process) executors.  Enable this option to achieve CPU affinity between ' +
    'acceptors and executors, instead of using underlying OS kernel scheduling algorithm.',
)


class Acceptor(multiprocessing.Process):
    """Work acceptor process.

    On start-up, `Acceptor` accepts a file descriptor which will be used to
    accept new work.  File descriptor is accepted over a `fd_queue`.

    `Acceptor` goes on to listen for new work over the received server socket.
    By default, `Acceptor` will spawn a new thread to handle each work.

    However, when ``--threadless`` option is enabled without ``--local-executor``,
    `Acceptor` process will also pre-spawns a
    :class:`~proxy.core.acceptor.threadless.Threadless` process during start-up.
    Accepted work is delegated to these :class:`~proxy.core.acceptor.threadless.Threadless`
    processes. `Acceptor` process shares accepted work with a
    :class:`~proxy.core.acceptor.threadless.Threadless` process over it's dedicated pipe.
    """

    def __init__(
            self,
            idd: int,
            fd_queue: connection.Connection,
            flags: argparse.Namespace,
            lock: 'multiprocessing.synchronize.Lock',
            # semaphore: multiprocessing.synchronize.Semaphore,
            executor_queues: List[connection.Connection],
            executor_pids: List[int],
            executor_locks: List['multiprocessing.synchronize.Lock'],
            event_queue: Optional[EventQueue] = None,
    ) -> None:
        super().__init__()
        self.flags = flags
        # Eventing core queue
        self.event_queue = event_queue
        # Index assigned by `AcceptorPool`
        self.idd = idd
        # Mutex used for synchronization with acceptors
        self.lock = lock
        # self.semaphore = semaphore
        # Queue over which server socket fd is received on start-up
        self.fd_queue: connection.Connection = fd_queue
        # Available executors
        self.executor_queues = executor_queues
        self.executor_pids = executor_pids
        self.executor_locks = executor_locks
        # Selector
        self.running = multiprocessing.Event()
        self.selector: Optional[selectors.DefaultSelector] = None
        # File descriptors used to accept new work
        self.socks: Dict[int, socket.socket] = {}
        # Internals
        self._total: Optional[int] = None
        self._local_work_queue: Optional['NonBlockingQueue'] = None
        self._local: Optional[LocalFdExecutor] = None
        self._lthread: Optional[threading.Thread] = None

    def accept(
            self,
            events: List[Tuple[selectors.SelectorKey, int]],
    ) -> List[Tuple[socket.socket, Optional[HostPort]]]:
        works = []
        for key, mask in events:
            if mask & selectors.EVENT_READ:
                try:
                    conn, addr = self.socks[key.data].accept()
                    logging.debug(
                        'Accepting new work#{0}'.format(conn.fileno()),
                    )
                    works.append((conn, addr or None))
                except BlockingIOError:
                    # logger.info('blocking io error')
                    pass
        return works

    def run_once(self) -> None:
        if self.selector is not None:
            events = self.selector.select(timeout=1)
            if len(events) == 0:
                return
            # locked = False
            # try:
            #     if self.lock.acquire(block=False):
            #         locked = True
            #         self.semaphore.release()
            # finally:
            #     if locked:
            #         self.lock.release()
            locked, works = False, []
            try:
                # if not self.semaphore.acquire(False, None):
                #     return
                if self.lock.acquire(block=False):
                    locked = True
                    works = self.accept(events)
            finally:
                if locked:
                    self.lock.release()
            for work in works:
                if self.flags.threadless and \
                        self.flags.local_executor:
                    assert self._local_work_queue
                    self._local_work_queue.put(work)
                else:
                    self._work(*work)

    def run(self) -> None:
        Logger.setup(
            self.flags.log_file, self.flags.log_level,
            self.flags.log_format,
        )
        self.selector = selectors.DefaultSelector()
        try:
            self._recv_and_setup_socks()
            if self.flags.threadless and self.flags.local_executor:
                self._start_local()
            for fileno in self.socks:
                self.selector.register(
                    fileno, selectors.EVENT_READ, fileno,
                )
            while not self.running.is_set():
                self.run_once()
        except KeyboardInterrupt:
            pass
        finally:
            for fileno in self.socks:
                self.selector.unregister(fileno)
            if self.flags.threadless and self.flags.local_executor:
                self._stop_local()
            for fileno in self.socks:
                self.socks[fileno].close()
            self.socks.clear()
            self.selector.close()
            logger.debug('Acceptor#%d shutdown', self.idd)

    def _recv_and_setup_socks(self) -> None:
        # TODO: Use selector on fd_queue so that we can
        # dynamically accept from new fds.
        for _ in range(self.fd_queue.recv()):
            fileno = recv_handle(self.fd_queue)
            sock = socket.socket(fileno=socket.dup(fileno))  # type: ignore[attr-defined]
            self.socks[fileno] = sock
        self.fd_queue.close()

    def _start_local(self) -> None:
        assert self.socks
        self._local_work_queue = NonBlockingQueue()
        self._local = LocalFdExecutor(
            iid=self.idd,
            work_queue=self._local_work_queue,
            flags=self.flags,
            event_queue=self.event_queue,
        )
        self._lthread = threading.Thread(target=self._local.run)
        self._lthread.daemon = True
        self._lthread.start()

    def _stop_local(self) -> None:
        if self._lthread is not None and \
                self._local_work_queue is not None:
            self._local_work_queue.put(False)
            self._lthread.join()

    def _work(self, conn: socket.socket, addr: Optional[HostPort]) -> None:
        self._total = self._total or 0
        if self.flags.threadless:
            # Index of worker to which this work should be dispatched
            # Use round-robin strategy by default.
            #
            # By default all acceptors will start sending work to
            # 1st workers.  To randomize, we offset index by idd.
            index = (self._total + self.idd) % self.flags.num_workers
            thread = threading.Thread(
                target=delegate_work_to_pool,
                args=(
                    self.executor_pids[index],
                    self.executor_queues[index],
                    self.executor_locks[index],
                    conn,
                    addr,
                    self.flags.unix_socket_path,
                ),
            )
            thread.start()
            # TODO: Move me into target method
            logger.debug(   # pragma: no cover
                'Dispatched work#{0}.{1}.{2} to worker#{3}'.format(
                    conn.fileno(), self.idd, self._total, index,
                ),
            )
        else:
            _, thread = start_threaded_work(
                self.flags,
                conn,
                addr,
                event_queue=self.event_queue,
                publisher_id=self.__class__.__qualname__,
            )
            # TODO: Move me into target method
            logger.debug(   # pragma: no cover
                'Started work#{0}.{1}.{2} in thread#{3}'.format(
                    conn.fileno(), self.idd, self._total, thread.ident,
                ),
            )
        self._total += 1



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.

    .. spelling::

       acceptor
       acceptors
       pre
"""
import logging
import argparse
import multiprocessing
from typing import TYPE_CHECKING, Any, List, Optional
from multiprocessing import connection
from multiprocessing.reduction import send_handle

from .acceptor import Acceptor
from ..listener import ListenerPool
from ...common.flag import flags
from ...common.constants import DEFAULT_NUM_ACCEPTORS


if TYPE_CHECKING:   # pragma: no cover
    from ..event import EventQueue

logger = logging.getLogger(__name__)


flags.add_argument(
    '--num-acceptors',
    type=int,
    default=DEFAULT_NUM_ACCEPTORS,
    help='Defaults to number of CPU cores.',
)


class AcceptorPool:
    """AcceptorPool is a helper class which pre-spawns
    :py:class:`~proxy.core.acceptor.acceptor.Acceptor` processes to
    utilize all available CPU cores for accepting new work.

    A file descriptor to consume work from is shared with
    :py:class:`~proxy.core.acceptor.acceptor.Acceptor` processes over a
    pipe.  Each :py:class:`~proxy.core.acceptor.acceptor.Acceptor`
    process then concurrently accepts new work over the shared file
    descriptor.

    Example usage:

        with AcceptorPool(flags=...) as pool:
            while True:
                time.sleep(1)

    `flags.work_klass` must implement :py:class:`~proxy.core.work.Work` class.
    """

    def __init__(
            self,
            flags: argparse.Namespace,
            listeners: ListenerPool,
            executor_queues: List[connection.Connection],
            executor_pids: List[int],
            executor_locks: List['multiprocessing.synchronize.Lock'],
            event_queue: Optional['EventQueue'] = None,
    ) -> None:
        self.flags = flags
        # File descriptor to use for accepting new work
        self.listeners: ListenerPool = listeners
        # Available executors
        self.executor_queues: List[connection.Connection] = executor_queues
        self.executor_pids: List[int] = executor_pids
        self.executor_locks: List['multiprocessing.synchronize.Lock'] = executor_locks
        # Eventing core queue
        self.event_queue: Optional['EventQueue'] = event_queue
        # Acceptor process instances
        self.acceptors: List[Acceptor] = []
        # Fd queues used to share file descriptor with acceptor processes
        self.fd_queues: List[connection.Connection] = []
        # Internals
        self.lock = multiprocessing.Lock()
        # self.semaphore = multiprocessing.Semaphore(0)

    def __enter__(self) -> 'AcceptorPool':
        self.setup()
        return self

    def __exit__(self, *args: Any) -> None:
        self.shutdown()

    def setup(self) -> None:
        """Setup acceptors."""
        self._start()
        execution_mode = (
            (
                'threadless (local)'
                if self.flags.local_executor
                else 'threadless (remote)'
            )
            if self.flags.threadless
            else 'threaded'
        )
        logger.debug(
            'Started %d acceptors in %s mode'
            % (
                self.flags.num_acceptors,
                execution_mode,
            ),
        )
        # Send file descriptor to all acceptor processes.
        for index in range(self.flags.num_acceptors):
            self.fd_queues[index].send(len(self.listeners.pool))
            for listener in self.listeners.pool:
                fd = listener.fileno()
                assert fd is not None
                send_handle(
                    self.fd_queues[index],
                    fd,
                    self.acceptors[index].pid,
                )
            self.fd_queues[index].close()

    def shutdown(self) -> None:
        logger.debug('Shutting down %d acceptors' % self.flags.num_acceptors)
        for acceptor in self.acceptors:
            acceptor.running.set()
        for acceptor in self.acceptors:
            acceptor.join()
        logger.debug('Acceptors shutdown')

    def _start(self) -> None:
        """Start acceptor processes."""
        for acceptor_id in range(self.flags.num_acceptors):
            work_queue = multiprocessing.Pipe()
            acceptor = Acceptor(
                idd=acceptor_id,
                fd_queue=work_queue[1],
                flags=self.flags,
                lock=self.lock,
                # semaphore=self.semaphore,
                event_queue=self.event_queue,
                executor_queues=self.executor_queues,
                executor_pids=self.executor_pids,
                executor_locks=self.executor_locks,
            )
            acceptor.start()
            logger.debug(
                'Started acceptor#%d process %d',
                acceptor_id,
                acceptor.pid,
            )
            self.acceptors.append(acceptor)
            self.fd_queues.append(work_queue[0])



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import struct
import logging
from typing import Tuple, Optional

from .hello import (
    TlsClientHello, TlsServerHello, TlsHelloRequest, TlsServerHelloDone,
)
from .types import tlsHandshakeType
from .finished import TlsFinished
from .certificate import (
    TlsCertificate, TlsCertificateVerify, TlsCertificateRequest,
)
from .key_exchange import TlsClientKeyExchange, TlsServerKeyExchange


logger = logging.getLogger(__name__)


class TlsHandshake:
    """TLS Handshake"""

    def __init__(self) -> None:
        self.msg_type: int = tlsHandshakeType.OTHER
        self.length: Optional[bytes] = None
        self.hello_request: Optional[TlsHelloRequest] = None
        self.client_hello: Optional[TlsClientHello] = None
        self.server_hello: Optional[TlsServerHello] = None
        self.certificate: Optional[TlsCertificate] = None
        self.server_key_exchange: Optional[TlsServerKeyExchange] = None
        self.certificate_request: Optional[TlsCertificateRequest] = None
        self.server_hello_done: Optional[TlsServerHelloDone] = None
        self.certificate_verify: Optional[TlsCertificateVerify] = None
        self.client_key_exchange: Optional[TlsClientKeyExchange] = None
        self.finished: Optional[TlsFinished] = None
        self.data: Optional[bytes] = None

    def parse(self, raw: bytes) -> Tuple[bool, bytes]:
        length = len(raw)
        if length < 4:
            logger.debug('invalid data, len(raw) = %s', length)
            return False, raw
        payload_length, = struct.unpack('!I', b'\x00' + raw[1:4])
        self.length = payload_length
        if length < 4 + payload_length:
            logger.debug(
                'incomplete data, len(raw) = %s, len(payload) = %s', length, payload_length,
            )
            return False, raw
        # parse
        self.msg_type = raw[0]
        self.length = raw[1:4]
        self.data = raw[: 4 + payload_length]
        payload = raw[4: 4 + payload_length]
        if self.msg_type == tlsHandshakeType.HELLO_REQUEST:
            # parse hello request
            self.hello_request = TlsHelloRequest()
            self.hello_request.parse(payload)
        elif self.msg_type == tlsHandshakeType.CLIENT_HELLO:
            # parse client hello
            self.client_hello = TlsClientHello()
            self.client_hello.parse(payload)
        elif self.msg_type == tlsHandshakeType.SERVER_HELLO:
            # parse server hello
            self.server_hello = TlsServerHello()
            self.server_hello.parse(payload)
        elif self.msg_type == tlsHandshakeType.CERTIFICATE:
            # parse certificate
            self.certificate = TlsCertificate()
            self.certificate.parse(payload)
        elif self.msg_type == tlsHandshakeType.SERVER_KEY_EXCHANGE:
            # parse server key exchange
            self.server_key_exchange = TlsServerKeyExchange()
            self.server_key_exchange.parse(payload)
        elif self.msg_type == tlsHandshakeType.CERTIFICATE_REQUEST:
            # parse certificate request
            self.certificate_request = TlsCertificateRequest()
            self.certificate_request.parse(payload)
        elif self.msg_type == tlsHandshakeType.SERVER_HELLO_DONE:
            # parse server hello done
            self.server_hello_done = TlsServerHelloDone()
            self.server_hello_done.parse(payload)
        elif self.msg_type == tlsHandshakeType.CERTIFICATE_VERIFY:
            # parse certificate verify
            self.certificate_verify = TlsCertificateVerify()
            self.certificate_verify.parse(payload)
        elif self.msg_type == tlsHandshakeType.CLIENT_KEY_EXCHANGE:
            # parse client key exchange
            self.client_key_exchange = TlsClientKeyExchange()
            self.client_key_exchange.parse(payload)
        elif self.msg_type == tlsHandshakeType.FINISHED:
            # parse finished
            self.finished = TlsFinished()
            self.finished.parse(payload)
        return True, raw[4 + payload_length:]

    def build(self) -> bytes:
        data = b''
        data += bytes([self.msg_type])
        payload = b''
        if self.msg_type == tlsHandshakeType.CLIENT_HELLO:
            assert self.client_hello
            payload = self.client_hello.build()
        elif self.msg_type == tlsHandshakeType.SERVER_HELLO:
            assert self.server_hello
            payload = self.server_hello.build()
        elif self.msg_type == tlsHandshakeType.CERTIFICATE:
            assert self.certificate
            payload = self.certificate.build()
        elif self.msg_type == tlsHandshakeType.SERVER_KEY_EXCHANGE:
            assert self.server_key_exchange
            payload = self.server_key_exchange.build()
        # calculate length
        length = struct.pack('!I', len(payload))[1:]
        data += length
        data += payload
        return data



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.

    .. spelling::

       http
       iterable
"""
from typing import NamedTuple


HttpStatusCodes = NamedTuple(
    'HttpStatusCodes', [
        # 1xx
        ('CONTINUE', int),
        ('SWITCHING_PROTOCOLS', int),
        # 2xx
        ('OK', int),
        # 3xx
        ('MOVED_PERMANENTLY', int),
        ('SEE_OTHER', int),
        ('TEMPORARY_REDIRECT', int),
        ('PERMANENT_REDIRECT', int),
        # 4xx
        ('BAD_REQUEST', int),
        ('UNAUTHORIZED', int),
        ('FORBIDDEN', int),
        ('NOT_FOUND', int),
        ('PROXY_AUTH_REQUIRED', int),
        ('REQUEST_TIMEOUT', int),
        ('I_AM_A_TEAPOT', int),
        # 5xx
        ('INTERNAL_SERVER_ERROR', int),
        ('NOT_IMPLEMENTED', int),
        ('BAD_GATEWAY', int),
        ('GATEWAY_TIMEOUT', int),
        ('NETWORK_READ_TIMEOUT_ERROR', int),
        ('NETWORK_CONNECT_TIMEOUT_ERROR', int),
    ],
)

httpStatusCodes = HttpStatusCodes(
    100, 101,
    200,
    301, 303, 307, 308,
    400, 401, 403, 404, 407, 408, 418,
    500, 501, 502, 504, 598, 599,
)



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
from .client import SocksClientConnection
from .packet import Socks4Packet
from .handler import SocksProtocolHandler
from .operations import Socks4Operations, socks4Operations


__all__ = [
    'Socks4Packet',
    'socks4Operations',
    'Socks4Operations',
    'SocksProtocolHandler',
    'SocksClientConnection',
]



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import socket
import argparse
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, List, Union, Optional

from .parser import HttpParser
from .connection import HttpClientConnection
from ..core.event import EventQueue
from .descriptors import DescriptorsHandlerMixin
from ..common.utils import tls_interception_enabled


if TYPE_CHECKING:   # pragma: no cover
    from ..core.connection import UpstreamConnectionPool


class HttpProtocolHandlerPlugin(
        DescriptorsHandlerMixin,
        ABC,
):
    """Base HttpProtocolHandler Plugin class.

    NOTE: This is an internal plugin and in most cases only useful for core contributors.
    If you are looking for proxy server plugins see `<proxy.HttpProxyBasePlugin>`.

    Implements various lifecycle events for an accepted client connection.
    Following events are of interest:

    1. Client Connection Accepted
       A new plugin instance is created per accepted client connection.
       Add your logic within __init__ constructor for any per connection setup.
    2. Client Request Chunk Received
       on_client_data is called for every chunk of data sent by the client.
    3. Client Request Complete
       on_request_complete is called once client request has completed.
    4. Server Response Chunk Received
       on_response_chunk is called for every chunk received from the server.
    5. Client Connection Closed
       Add your logic within `on_client_connection_close` for any per connection tear-down.
    """

    def __init__(
            self,
            uid: str,
            flags: argparse.Namespace,
            client: HttpClientConnection,
            request: HttpParser,
            event_queue: Optional[EventQueue] = None,
            upstream_conn_pool: Optional['UpstreamConnectionPool'] = None,
    ):
        self.uid: str = uid
        self.flags: argparse.Namespace = flags
        self.client: HttpClientConnection = client
        self.request: HttpParser = request
        self.event_queue = event_queue
        self.upstream_conn_pool = upstream_conn_pool

    @staticmethod
    @abstractmethod
    def protocols() -> List[int]:
        raise NotImplementedError()

    @abstractmethod
    def on_client_data(self, raw: memoryview) -> None:
        """Called only after original request has been completely received."""
        pass    # pragma: no cover

    @abstractmethod
    def on_request_complete(self) -> Union[socket.socket, bool]:
        """Called right after client request parser has reached COMPLETE state."""
        return False  # pragma: no cover

    @abstractmethod
    def on_response_chunk(self, chunk: List[memoryview]) -> List[memoryview]:
        """Handle data chunks as received from the server.

        Return optionally modified chunk to return back to client."""
        return chunk  # pragma: no cover

    @abstractmethod
    def on_client_connection_close(self) -> None:
        """Client connection shutdown has been received, flush has been called,
        perform any cleanup work here.
        """
        pass  # pragma: no cover

    @property
    def tls_interception_enabled(self) -> bool:
        return tls_interception_enabled(self.flags)



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import logging
import argparse
from typing import TYPE_CHECKING

from .base import BaseSshTunnelHandler


logger = logging.getLogger(__name__)

if TYPE_CHECKING:   # pragma: no cover
    from ...common.types import HostPort
    try:
        from paramiko.channel import Channel
    except ImportError:
        pass


class SshHttpProtocolHandler(BaseSshTunnelHandler):
    """Handles incoming connections over forwarded SSH transport."""

    def __init__(self, flags: argparse.Namespace) -> None:
        self.flags = flags

    def on_connection(
            self,
            chan: 'Channel',
            origin: 'HostPort',
            server: 'HostPort',
    ) -> None:
        logger.debug('handle proxy request')

    def shutdown(self) -> None:
        logger.debug('ssh handler shutdown')



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import argparse
from abc import ABC
from typing import TYPE_CHECKING, Any, Dict, Tuple, Optional

from ..parser import HttpParser
from ..connection import HttpClientConnection
from ...core.event import EventQueue
from ..descriptors import DescriptorsHandlerMixin
from ...common.utils import tls_interception_enabled


if TYPE_CHECKING:   # pragma: no cover
    from ...common.types import HostPort
    from ...core.connection import UpstreamConnectionPool


class HttpProxyBasePlugin(
        DescriptorsHandlerMixin,
        ABC,
):
    """Base HttpProxyPlugin Plugin class.

    Implement various lifecycle event methods to customize behavior."""

    def __init__(
            self,
            uid: str,
            flags: argparse.Namespace,
            client: HttpClientConnection,
            event_queue: EventQueue,
            upstream_conn_pool: Optional['UpstreamConnectionPool'] = None,
    ) -> None:
        self.uid = uid                  # pragma: no cover
        self.flags = flags              # pragma: no cover
        self.client = client            # pragma: no cover
        self.event_queue = event_queue  # pragma: no cover
        self.upstream_conn_pool = upstream_conn_pool

    def name(self) -> str:
        """A unique name for your plugin.

        Defaults to name of the class. This helps plugin developers to directly
        access a specific plugin by its name."""
        return self.__class__.__qualname__      # pragma: no cover

    def resolve_dns(self, host: str, port: int) -> Tuple[Optional[str], Optional['HostPort']]:
        """Resolve upstream server host to an IP address.

        Optionally also override the source address to use for
        connection with upstream server.

        For upstream IP:
        Return None to use default resolver available to the system.
        Return IP address as string to use your custom resolver.

        For source address:
        Return None to use default source address
        Return 2-tuple representing (host, port) to use as source address
        """
        return None, None

    # No longer abstract since 2.4.0
    #
    # @abstractmethod
    def before_upstream_connection(
            self, request: HttpParser,
    ) -> Optional[HttpParser]:
        """Handler called just before Proxy upstream connection is established.

        Return optionally modified request object.
        If None is returned, upstream connection won't be established.

        Raise HttpRequestRejected or HttpProtocolException directly to drop the connection."""
        return request  # pragma: no cover

    # Since 3.4.0
    #
    # @abstractmethod
    def handle_client_data(
            self, raw: memoryview,
    ) -> Optional[memoryview]:
        """Handler called in special scenarios when an upstream server connection
        is never established.

        Essentially, if you return None from within before_upstream_connection,
        be prepared to handle_client_data and not handle_client_request.

        Only called after initial request from client has been received.

        Raise HttpRequestRejected to tear down the connection
        Return None to drop the connection
        """
        return raw  # pragma: no cover

    # No longer abstract since 2.4.0
    #
    # @abstractmethod
    def handle_client_request(
            self, request: HttpParser,
    ) -> Optional[HttpParser]:
        """Handler called before dispatching client request to upstream.

        Note: For pipelined (keep-alive) connections, this handler can be
        called multiple times, for each request sent to upstream.

        Note: If TLS interception is enabled, this handler can
        be called multiple times if client exchanges multiple
        requests over same SSL session.

        Return optionally modified request object to dispatch to upstream.
        Return None to drop the request data, e.g. in case a response has already been queued.
        Raise HttpRequestRejected or HttpProtocolException directly to
        tear down the connection with client.
        """
        return request  # pragma: no cover

    # No longer abstract since 2.4.0
    #
    # @abstractmethod
    def handle_upstream_chunk(self, chunk: memoryview) -> Optional[memoryview]:
        """Handler called right after receiving raw response from upstream server.

        For HTTPS connections, chunk will be encrypted unless
        TLS interception is also enabled.

        Return None if you don't want to sent this chunk to the client.
        """
        return chunk  # pragma: no cover

    # No longer abstract since 2.4.0
    #
    # @abstractmethod
    def on_upstream_connection_close(self) -> None:
        """Handler called right after upstream connection has been closed."""
        pass  # pragma: no cover

    def on_access_log(self, context: Dict[str, Any]) -> Optional[Dict[str, Any]]:
        """Use this method to override default access log format (see
        DEFAULT_HTTP_ACCESS_LOG_FORMAT and DEFAULT_HTTPS_ACCESS_LOG_FORMAT) and to
        add/update/modify/delete context for next plugin.on_access_log invocation.

        This is specially useful if a plugins want to provide extra context
        in the access log which may not available within other plugins' context or even
        in proxy.py core.

        Returns Log context or None.  If plugin chooses to access log, they ideally
        must return None to prevent other plugin.on_access_log invocation.
        """
        return context

    def do_intercept(self, _request: HttpParser) -> bool:
        """By default returns True (only) when necessary flags
        for TLS interception are passed.

        When TLS interception is enabled, plugins can still disable
        TLS interception by returning False explicitly.  This hook
        will allow you to run proxy instance with TLS interception
        flags BUT only conditionally enable interception for
        certain requests.
        """
        return tls_interception_enabled(self.flags)



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.

    .. spelling::

       utils
"""
import os
import ssl
import sys
import socket
import logging
import argparse
import tempfile
import functools
import ipaddress
import contextlib
from types import TracebackType
from typing import Any, Dict, List, Type, Tuple, Callable, Optional

from .types import HostPort
from .constants import (
    CRLF, COLON, HTTP_1_1, IS_WINDOWS, WHITESPACE, DEFAULT_TIMEOUT,
    DEFAULT_THREADLESS, DEFAULT_CONNECT_TIMEOUT, PROXY_AGENT_HEADER_VALUE,
    DEFAULT_SSL_CONTEXT_OPTIONS,
)


if not IS_WINDOWS:  # pragma: no cover
    import resource

logger = logging.getLogger(__name__)


def cert_der_to_dict(der: Optional[bytes]) -> Dict[str, Any]:
    """Parse a DER formatted certificate to a python dict"""
    # pylint: disable=import-outside-toplevel
    import _ssl  # noqa: WPS436

    if not der:
        return {}
    with tempfile.NamedTemporaryFile(delete=False) as cert_file:
        pem = ssl.DER_cert_to_PEM_cert(der)
        cert_file.write(pem.encode())
        cert_file.flush()
        cert_file.seek(0)
    try:
        certificate = _ssl._test_decode_cert(cert_file.name)
    finally:
        cert_file.close()
        os.remove(cert_file.name)
    return certificate or {}


def tls_interception_enabled(flags: argparse.Namespace) -> bool:
    return flags.ca_key_file is not None and \
        flags.ca_cert_dir is not None and \
        flags.ca_signing_key_file is not None and \
        flags.ca_cert_file is not None


def is_threadless(threadless: bool, threaded: bool) -> bool:
    # if default is threadless then return true unless
    # user has overridden mode using threaded flag.
    #
    # if default is not threadless then return true
    # only if user has overridden using --threadless flag
    return (DEFAULT_THREADLESS and not threaded) or (not DEFAULT_THREADLESS and threadless)


def is_py2() -> bool:
    """Exists only to avoid mocking :data:`sys.version_info` in tests."""
    return sys.version_info.major == 2


def text_(s: Any, encoding: str = 'utf-8', errors: str = 'strict') -> Any:
    """Utility to ensure text-like usability.

    If s is of type bytes or int, return s.decode(encoding, errors),
    otherwise return s as it is."""
    if isinstance(s, int):
        return str(s)
    if isinstance(s, bytes):
        return s.decode(encoding, errors)
    return s


def bytes_(s: Any, encoding: str = 'utf-8', errors: str = 'strict') -> Any:
    """Utility to ensure binary-like usability.

    If s is type str or int, return s.encode(encoding, errors),
    otherwise return s as it is."""
    if isinstance(s, int):
        s = str(s)
    if isinstance(s, str):
        return s.encode(encoding, errors)
    return s


def build_http_request(
    method: bytes, url: bytes,
    protocol_version: bytes = HTTP_1_1,
    content_type: Optional[bytes] = None,
    headers: Optional[Dict[bytes, bytes]] = None,
    body: Optional[bytes] = None,
    conn_close: bool = False,
    no_ua: bool = False,
) -> bytes:
    """Build and returns a HTTP request packet."""
    headers = headers or {}
    if content_type is not None:
        headers[b'Content-Type'] = content_type
    has_transfer_encoding = False
    has_user_agent = False
    for k, _ in headers.items():
        if k.lower() == b'transfer-encoding':
            has_transfer_encoding = True
        elif k.lower() == b'user-agent':
            has_user_agent = True
    if body and not has_transfer_encoding:
        headers[b'Content-Length'] = bytes_(len(body))
    if not has_user_agent and not no_ua:
        headers[b'User-Agent'] = PROXY_AGENT_HEADER_VALUE
    return build_http_pkt(
        [method, url, protocol_version],
        headers,
        body,
        conn_close,
    )


def build_http_response(
    status_code: int,
    protocol_version: bytes = HTTP_1_1,
    reason: Optional[bytes] = None,
    headers: Optional[Dict[bytes, bytes]] = None,
    body: Optional[bytes] = None,
    conn_close: bool = False,
    no_cl: bool = False,
) -> bytes:
    """Build and returns a HTTP response packet."""
    line = [protocol_version, bytes_(status_code)]
    if reason:
        line.append(reason)
    headers = headers or {}
    has_transfer_encoding = False
    for k, _ in headers.items():
        if k.lower() == b'transfer-encoding':
            has_transfer_encoding = True
            break
    if not has_transfer_encoding and not no_cl:
        headers[b'Content-Length'] = bytes_(len(body)) if body else b'0'
    return build_http_pkt(line, headers, body, conn_close)


def build_http_header(k: bytes, v: bytes) -> bytes:
    """Build and return a HTTP header line for use in raw packet."""
    return k + COLON + WHITESPACE + v


def build_http_pkt(
    line: List[bytes],
    headers: Optional[Dict[bytes, bytes]] = None,
    body: Optional[bytes] = None,
    conn_close: bool = False,
) -> bytes:
    """Build and returns a HTTP request or response packet."""
    pkt = WHITESPACE.join(line) + CRLF
    headers = headers or {}
    if conn_close:
        headers[b'Connection'] = b'close'
    for k, v in headers.items():
        pkt += build_http_header(k, v) + CRLF
    pkt += CRLF
    if body:
        pkt += body
    return pkt


def build_websocket_handshake_request(
        key: bytes,
        method: bytes = b'GET',
        url: bytes = b'/',
        host: bytes = b'localhost',
) -> bytes:
    """
    Build and returns a Websocket handshake request packet.

    :param key: Sec-WebSocket-Key header value.
    :param method: HTTP method.
    :param url: Websocket request path.
    """
    return build_http_request(
        method, url,
        headers={
            b'Host': host,
            b'Connection': b'upgrade',
            b'Upgrade': b'websocket',
            b'Sec-WebSocket-Key': key,
            b'Sec-WebSocket-Version': b'13',
        },
    )


def build_websocket_handshake_response(accept: bytes) -> bytes:
    """
    Build and returns a Websocket handshake response packet.

    :param accept: Sec-WebSocket-Accept header value
    """
    return build_http_response(
        101, reason=b'Switching Protocols',
        headers={
            b'Upgrade': b'websocket',
            b'Connection': b'Upgrade',
            b'Sec-WebSocket-Accept': accept,
        },
    )


def find_http_line(raw: bytes) -> Tuple[Optional[bytes], bytes]:
    """Find and returns first line ending in CRLF along with following buffer.

    If no ending CRLF is found, line is None."""
    parts = raw.split(CRLF, 1)
    return (None, raw) \
        if len(parts) == 1 \
        else (parts[0], parts[1])


def wrap_socket(
        conn: socket.socket,
        keyfile: str,
        certfile: str,
        cafile: Optional[str] = None,
) -> ssl.SSLSocket:
    """Use this to upgrade server_side socket to TLS."""
    ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH, cafile=cafile)
    ctx.options |= DEFAULT_SSL_CONTEXT_OPTIONS
    ctx.verify_mode = ssl.CERT_NONE
    ctx.load_cert_chain(certfile=certfile, keyfile=keyfile)
    return ctx.wrap_socket(conn, server_side=True)


def new_socket_connection(
    addr: HostPort,
    timeout: float = DEFAULT_TIMEOUT,
    connect_timeout: float = DEFAULT_CONNECT_TIMEOUT,
    source_address: Optional[HostPort] = None,
) -> socket.socket:
    conn = None
    try:
        ip = ipaddress.ip_address(addr[0])
        if ip.version == 4:
            conn = socket.socket(
                socket.AF_INET, socket.SOCK_STREAM, 0,
            )
            conn.settimeout(connect_timeout)
            conn.connect(addr)
        else:
            conn = socket.socket(
                socket.AF_INET6, socket.SOCK_STREAM, 0,
            )
            conn.settimeout(connect_timeout)
            conn.connect((addr[0], addr[1], 0, 0))
    except ValueError:
        pass    # does not appear to be an IPv4 or IPv6 address

    if conn is not None:
        conn.settimeout(timeout)
        return conn

    # try to establish dual stack IPv4/IPv6 connection.
    conn = socket.create_connection(
        addr,
        timeout=connect_timeout,
        source_address=source_address,
    )
    conn.settimeout(timeout)
    return conn


class socket_connection(contextlib.ContextDecorator):
    """Same as new_socket_connection but as a context manager and decorator."""

    def __init__(self, addr: HostPort):
        self.addr: HostPort = addr
        self.conn: Optional[socket.socket] = None
        super().__init__()

    def __enter__(self) -> socket.socket:
        self.conn = new_socket_connection(self.addr)
        return self.conn

    def __exit__(
            self,
            exc_type: Optional[Type[BaseException]],
            exc_val: Optional[BaseException],
            exc_tb: Optional[TracebackType],
    ) -> None:
        if self.conn:
            self.conn.close()

    def __call__(   # type: ignore
            self, func: Callable[..., Any],
    ) -> Callable[[Tuple[Any, ...], Dict[str, Any]], Any]:
        @functools.wraps(func)
        def decorated(*args: Any, **kwargs: Any) -> Any:
            with self as conn:
                return func(conn, *args, **kwargs)
        return decorated


def get_available_port() -> int:
    """Finds and returns an available port on the system."""
    with contextlib.closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as sock:
        sock.bind(('', 0))
        _, port = sock.getsockname()
    return int(port)


def set_open_file_limit(soft_limit: int) -> None:
    """Configure open file description soft limit on supported OS."""
    # resource module not available on Windows OS
    if IS_WINDOWS:  # pragma: no cover
        return

    # pylint: disable=possibly-used-before-assignment
    curr_soft_limit, curr_hard_limit = resource.getrlimit(
        resource.RLIMIT_NOFILE,
    )
    if curr_soft_limit < soft_limit < curr_hard_limit:
        resource.setrlimit(
            resource.RLIMIT_NOFILE, (soft_limit, curr_hard_limit),
        )
        logger.debug(
            'Open file soft limit set to %d', soft_limit,
        )



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.

    .. spelling::

       reusability
"""
import socket
import logging
import selectors
from typing import TYPE_CHECKING, Any, Set, Dict, Tuple

from ..work import Work
from .server import TcpServerConnection
from ...common.flag import flags
from ...common.types import HostPort, Readables, Writables, SelectableEvents


logger = logging.getLogger(__name__)


flags.add_argument(
    '--enable-conn-pool',
    action='store_true',
    default=False,
    help='Default: False.  (WIP) Enable upstream connection pooling.',
)


class UpstreamConnectionPool(Work[TcpServerConnection]):
    """Manages connection pool to upstream servers.

    `UpstreamConnectionPool` avoids need to reconnect with the upstream
    servers repeatedly when a reusable connection is available
    in the pool.

    A separate pool is maintained for each upstream server.
    So internally, it's a pool of pools.

    Internal data structure maintains references to connection objects
    that pool owns or has borrowed.  Borrowed connections are marked as
    NOT reusable.

    For reusable connections only, pool listens for read events
    to detect broken connections.  This can happen if pool has opened
    a connection, which was never used and eventually reaches
    upstream server timeout limit.

    When a borrowed connection is returned back to the pool,
    the connection is marked as reusable again.  However, if
    returned connection has already been closed, it is removed
    from the internal data structure.

    TODO: Ideally, `UpstreamConnectionPool` must be shared across
    all cores to make SSL session cache to also work
    without additional out-of-bound synchronizations.

    TODO: `UpstreamConnectionPool` currently WON'T work for
    HTTPS connection. This is because of missing support for
    session cache, session ticket, abbr TLS handshake
    and other necessary features to make it work.

    NOTE: However, currently for all HTTP only upstream connections,
    `UpstreamConnectionPool` can be used to remove slow starts.
    """

    def __init__(self) -> None:
        self.connections: Dict[int, TcpServerConnection] = {}
        self.pools: Dict[HostPort, Set[TcpServerConnection]] = {}

    @staticmethod
    def create(*args: Any) -> TcpServerConnection:   # pragma: no cover
        return TcpServerConnection(*args)

    def acquire(self, addr: HostPort) -> Tuple[bool, TcpServerConnection]:
        """Returns a reusable connection from the pool.

        If none exists, will create and return a new connection."""
        created, conn = False, None
        if addr in self.pools:
            for old_conn in self.pools[addr]:
                if old_conn.is_reusable():
                    conn = old_conn
                    logger.debug(
                        'Reusing connection#{2} for upstream {0}:{1}'.format(
                            addr[0], addr[1], id(old_conn),
                        ),
                    )
                    break
        if conn is None:
            created, conn = True, self.add(addr)
        conn.mark_inuse()
        return created, conn

    def release(self, conn: TcpServerConnection) -> None:
        """Release a previously acquired connection.

        Releasing a connection will shutdown and close the socket
        including internal pool cleanup.
        """
        assert not conn.is_reusable()
        logger.debug(
            'Removing connection#{2} from pool from upstream {0}:{1}'.format(
                conn.addr[0], conn.addr[1], id(conn),
            ),
        )
        self._remove(conn.connection.fileno())

    def retain(self, conn: TcpServerConnection) -> None:
        """Retained previously acquired connection in the pool for reusability."""
        assert not conn.closed
        logger.debug(
            'Retaining connection#{2} to upstream {0}:{1}'.format(
                conn.addr[0], conn.addr[1], id(conn),
            ),
        )
        # Reset for reusability
        conn.reset()

    async def get_events(self) -> SelectableEvents:
        """Returns read event flag for all reusable connections in the pool."""
        events = {}
        for connections in self.pools.values():
            for conn in connections:
                if conn.is_reusable():
                    events[conn.connection.fileno()] = selectors.EVENT_READ
        return events

    async def handle_events(self, readables: Readables, _writables: Writables) -> bool:
        """Removes reusable connection from the pool.

        When pool is the owner of connection, we don't expect a read event from upstream
        server.  A read event means either upstream closed the connection or connection
        has somehow reached an illegal state e.g. upstream sending data for previous
        connection acquisition lifecycle."""
        for fileno in readables:
            if TYPE_CHECKING:   # pragma: no cover
                assert isinstance(fileno, int)
            logger.debug('Upstream fd#{0} is read ready'.format(fileno))
            self._remove(fileno)
        return False

    def add(self, addr: HostPort) -> TcpServerConnection:
        """Creates, connects and adds a new connection to the pool.

        Returns newly created connection.

        NOTE: You must not use the returned connection, instead use `acquire`.
        """
        new_conn = self.create(addr[0], addr[1])
        new_conn.connect()
        self._add(new_conn)
        logger.debug(
            'Created new connection#{2} for upstream {0}:{1}'.format(
                addr[0], addr[1], id(new_conn),
            ),
        )
        return new_conn

    def _add(self, conn: TcpServerConnection) -> None:
        """Adds a new connection to internal data structure."""
        if conn.addr not in self.pools:
            self.pools[conn.addr] = set()
        conn._reusable = True
        self.pools[conn.addr].add(conn)
        self.connections[conn.connection.fileno()] = conn

    def _remove(self, fileno: int) -> None:
        """Remove a connection by descriptor from the internal data structure."""
        conn = self.connections[fileno]
        logger.debug('Removing conn#{0} from pool'.format(id(conn)))
        try:
            conn.connection.shutdown(socket.SHUT_WR)
        except OSError:
            pass
        conn.close()
        self.pools[conn.addr].remove(conn)
        del self.connections[fileno]



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
from .url import Url
from .codes import httpStatusCodes
from .plugin import HttpProtocolHandlerPlugin
from .handler import HttpProtocolHandler
from .headers import httpHeaders
from .methods import httpMethods
from .connection import HttpClientConnection


__all__ = [
    'HttpProtocolHandler',
    'HttpClientConnection',
    'HttpProtocolHandlerPlugin',
    'httpStatusCodes',
    'httpMethods',
    'httpHeaders',
    'Url',
]



# This stub file is necessary because `_scm_version.py`
# autogenerated on build and absent on mypy checks time
from typing import Tuple, Union


version: str
version_tuple: Tuple[Union[int, str], ...]



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import socket
import logging
import argparse
from abc import ABC, abstractmethod
from typing import Any, Optional

from ...common.flag import flags
from ...common.constants import DEFAULT_BACKLOG


flags.add_argument(
    '--backlog',
    type=int,
    default=DEFAULT_BACKLOG,
    help='Default: 100. Maximum number of pending connections to proxy server.',
)

logger = logging.getLogger(__name__)


class BaseListener(ABC):
    """Base listener class.

    For usage provide a listen method implementation."""

    def __init__(self, *args: Any, flags: argparse.Namespace, **kwargs: Any) -> None:
        self.flags = flags
        self._socket: Optional[socket.socket] = None

    @abstractmethod
    def listen(self) -> socket.socket:
        raise NotImplementedError()

    def __enter__(self) -> 'BaseListener':
        self.setup()
        return self

    def __exit__(self, *args: Any) -> None:
        self.shutdown()

    def fileno(self) -> Optional[int]:
        if not self._socket:
            return None
        return self._socket.fileno()

    def setup(self) -> None:
        self._socket = self.listen()

    def shutdown(self) -> None:
        assert self._socket
        self._socket.close()



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
from typing import Tuple, Optional

from ..exception import HttpProtocolException
from ...common.constants import WHITESPACE


PROXY_PROTOCOL_V2_SIGNATURE = b'\x0D\x0A\x0D\x0A\x00\x0D\x0A\x51\x55\x49\x54\x0A'


class ProxyProtocol:
    """Reference https://www.haproxy.org/download/1.8/doc/proxy-protocol.txt"""

    def __init__(self) -> None:
        self.version: Optional[int] = None
        self.family: Optional[bytes] = None
        self.source: Optional[Tuple[bytes, int]] = None
        self.destination: Optional[Tuple[bytes, int]] = None

    def parse(self, raw: bytes) -> None:
        if raw.startswith(b'PROXY'):
            self.version = 1
            # Per spec, v1 line cannot exceed this limit
            assert len(raw) <= 57
            line = raw.split(WHITESPACE)
            assert line[0] == b'PROXY' and line[1] in (
                b'TCP4', b'TCP6', b'UNKNOWN',
            )
            self.family = line[1]
            if len(line) == 6:
                self.source = (line[2], int(line[4]))
                self.destination = (line[3], int(line[5]))
            else:
                assert self.family == b'UNKNOWN'
        elif raw.startswith(PROXY_PROTOCOL_V2_SIGNATURE):
            self.version = 2
            raise NotImplementedError()
        else:
            raise HttpProtocolException(
                'Neither a v1 or v2 proxy protocol packet',
            )



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import time
import uuid
import threading
from typing import Any

from ..local import BaseLocalExecutor
from ....common.backports import NonBlockingQueue


class LocalTaskExecutor(BaseLocalExecutor):
    """We'll define a local executor which is capable of receiving
    log lines over a non blocking queue."""

    def work(self, *args: Any) -> None:
        task_id = int(time.time())
        uid = '%s-%s' % (self.iid, task_id)
        self.works[task_id] = self.create(uid, *args)


class ThreadedTaskExecutor(threading.Thread):

    def __init__(self, **kwargs: Any) -> None:
        super().__init__()
        self.daemon = True
        self.executor = LocalTaskExecutor(
            iid=uuid.uuid4().hex,
            work_queue=NonBlockingQueue(),
            **kwargs,
        )

    def __enter__(self) -> 'ThreadedTaskExecutor':
        self.start()
        return self

    def __exit__(self, *args: Any) -> None:
        self.executor.running.set()
        self.join()

    def run(self) -> None:
        self.executor.run()



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import logging
from typing import Any, Optional

from .store.base import CacheStore
from ...http.proxy import HttpProxyBasePlugin
from ...common.flag import flags
from ...http.parser import HttpParser
from ...common.constants import (
    DEFAULT_CACHE_REQUESTS, DEFAULT_CACHE_BY_CONTENT_TYPE,
)


logger = logging.getLogger(__name__)


flags.add_argument(
    '--cache-requests',
    action='store_true',
    default=DEFAULT_CACHE_REQUESTS,
    help='Default: False.  ' +
    'Whether to also write request packets in the cache file.',
)

flags.add_argument(
    '--cache-by-content-type',
    action='store_true',
    default=DEFAULT_CACHE_BY_CONTENT_TYPE,
    help='Default: False.  ' +
    'Whether to extract content by type from responses. ' +
    'Extracted content type is written to the cache directory e.g. video.mp4.',
)


class BaseCacheResponsesPlugin(HttpProxyBasePlugin):
    """Base cache plugin.

    It requires a storage backend to work with. Storage class
    must implement CacheStore interface.

    Different storage backends can be used per request if required.
    """

    def __init__(
            self,
            *args: Any,
            **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.store: Optional[CacheStore] = None

    def set_store(self, store: CacheStore) -> None:
        self.store = store

    def before_upstream_connection(
            self, request: HttpParser,
    ) -> Optional[HttpParser]:
        assert self.store
        try:
            self.store.open(request)
        except Exception as e:
            logger.info('Caching disabled due to exception message %s', str(e))
        return request

    def handle_client_request(
            self, request: HttpParser,
    ) -> Optional[HttpParser]:
        assert self.store
        return self.store.cache_request(request)

    def handle_upstream_chunk(self, chunk: memoryview) -> Optional[memoryview]:
        assert self.store
        return self.store.cache_response_chunk(chunk)

    def on_upstream_connection_close(self) -> None:
        assert self.store
        self.store.close()



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import logging
from abc import ABC, abstractmethod
from typing import List, Union, Optional

from .types import tcpConnectionTypes
from ...common.types import TcpOrTlsSocket
from ...common.constants import DEFAULT_BUFFER_SIZE, DEFAULT_MAX_SEND_SIZE


logger = logging.getLogger(__name__)


class TcpConnectionUninitializedException(Exception):
    pass


class TcpConnection(ABC):
    """TCP server/client connection abstraction.

    Main motivation of this class is to provide a buffer management
    when reading and writing into the socket.

    Implement the connection property abstract method to return
    a socket connection object.
    """

    def __init__(self, tag: int) -> None:
        self.tag: str = 'server' if tag == tcpConnectionTypes.SERVER else 'client'
        self.buffer: List[memoryview] = []
        self.closed: bool = False
        self._reusable: bool = False
        self._num_buffer = 0

    @property
    @abstractmethod
    def connection(self) -> TcpOrTlsSocket:
        """Must return the socket connection to use in this class."""
        raise TcpConnectionUninitializedException()     # pragma: no cover

    def send(self, data: Union[memoryview, bytes]) -> int:
        """Users must handle BrokenPipeError exceptions"""
        return self.connection.send(data)

    def recv(
            self, buffer_size: int = DEFAULT_BUFFER_SIZE,
    ) -> Optional[memoryview]:
        """Users must handle socket.error exceptions"""
        data: bytes = self.connection.recv(buffer_size)
        if len(data) == 0:
            return None
        logger.debug(
            'received %d bytes from %s' %
            (len(data), self.tag),
        )
        # logger.info(data)
        return memoryview(data)

    def close(self) -> bool:
        if not self.closed and self.connection:
            self.connection.close()
            self.closed = True
        return self.closed

    def has_buffer(self) -> bool:
        return self._num_buffer != 0

    def queue(self, mv: memoryview) -> None:
        self.buffer.append(mv)
        self._num_buffer += 1

    def flush(self, max_send_size: Optional[int] = None) -> int:
        """Users must handle BrokenPipeError exceptions"""
        if not self.has_buffer():
            return 0
        mv = self.buffer[0]
        # TODO: Assemble multiple packets if total
        # size remains below max send size.
        max_send_size = max_send_size or DEFAULT_MAX_SEND_SIZE
        try:
            sent: int = self.send(mv[:max_send_size])
        except BlockingIOError:
            logger.warning('BlockingIOError when trying send to {0}'.format(self.tag))
            return 0
        if sent == len(mv):
            self.buffer.pop(0)
            self._num_buffer -= 1
        else:
            self.buffer[0] = mv[sent:]
        logger.debug('flushed %d bytes to %s' % (sent, self.tag))
        # logger.info(mv[:sent].tobytes())
        del mv
        return sent

    def is_reusable(self) -> bool:
        return self._reusable

    def mark_inuse(self) -> None:
        self._reusable = False

    def reset(self) -> None:
        assert not self.closed
        self._reusable = True
        self.buffer = []
        self._num_buffer = 0



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import ssl
from typing import Any, Dict, Optional

from .types import tcpConnectionTypes
from .connection import TcpConnection, TcpConnectionUninitializedException
from ...common.types import HostPort, TcpOrTlsSocket
from ...common.constants import DEFAULT_SSL_CONTEXT_OPTIONS


class TcpClientConnection(TcpConnection):
    """A buffered client connection object."""

    def __init__(
        self,
        conn: TcpOrTlsSocket,
        # optional for unix socket servers
        addr: Optional[HostPort] = None,
    ) -> None:
        super().__init__(tcpConnectionTypes.CLIENT)
        self._conn: Optional[TcpOrTlsSocket] = conn
        self.addr: Optional[HostPort] = addr

    @property
    def address(self) -> str:
        return 'unix:client' if not self.addr else '{0}:{1}'.format(self.addr[0], self.addr[1])

    @property
    def connection(self) -> TcpOrTlsSocket:
        if self._conn is None:
            raise TcpConnectionUninitializedException()
        return self._conn

    def wrap(self, keyfile: str, certfile: str) -> None:
        self.connection.setblocking(True)
        self.flush()
        ctx = ssl.SSLContext(
            protocol=(
                ssl.PROTOCOL_TLS_CLIENT
                if self.tag == 'server'
                else ssl.PROTOCOL_TLS_SERVER
            ),
        )
        ctx.options |= DEFAULT_SSL_CONTEXT_OPTIONS
        ctx.load_cert_chain(certfile=certfile, keyfile=keyfile)
        assert self.addr
        kwargs: Dict[str, Any] = {'server_side': True}
        if self.tag == 'server':
            assert self.addr
            kwargs['server_hostname'] = self.addr[0]
        self._conn = ctx.wrap_socket(self.connection, **kwargs)
        self.connection.setblocking(False)



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import uuid
import queue
import logging
import threading
import multiprocessing
from typing import Any, Dict, Callable, Optional
from multiprocessing import connection

from .names import eventNames
from .queue import EventQueue


logger = logging.getLogger(__name__)


class EventSubscriber:
    """Core event subscriber.

    Usage: Initialize one instance per CPU core for optimum performance.

    EventSubscriber can run within various context. E.g. main thread,
    another thread or a different process.  EventSubscriber context
    can be different from publishers.  Publishers can even be processes
    outside of the proxy.py core.

    `multiprocessing.Pipe` is used to initialize a new Queue for
    receiving subscribed events from eventing core.  Note that,
    core EventDispatcher might be running in a separate process
    and hence subscription queue must be multiprocess safe.

    When `subscribe` method is called, EventManager stars
    a relay thread which consumes event out of the subscription queue
    and invoke callback.

    NOTE: Callback is executed in the context of relay thread.
    """

    def __init__(self, event_queue: EventQueue, callback: Callable[[Dict[str, Any]], None]) -> None:
        self.event_queue = event_queue
        self.callback = callback
        self.relay_thread: Optional[threading.Thread] = None
        self.relay_shutdown: Optional[threading.Event] = None
        self.relay_recv: Optional[connection.Connection] = None
        self.relay_send: Optional[connection.Connection] = None
        self.relay_sub_id: Optional[str] = None

    def __enter__(self) -> 'EventSubscriber':
        self.setup()
        return self

    def __exit__(self, *args: Any) -> None:
        self.shutdown()

    def setup(self, do_subscribe: bool = True) -> None:
        """Setup subscription thread.

        Call subscribe() to actually start subscription.
        """
        self._start_relay_thread()
        assert self.relay_sub_id and self.relay_recv
        logger.debug(
            'Subscriber#%s relay setup done',
            self.relay_sub_id,
        )
        if do_subscribe:
            self.subscribe()

    def shutdown(self, do_unsubscribe: bool = True) -> None:
        """Tear down subscription thread.

        Call unsubscribe() to actually stop subscription.
        """
        self._stop_relay_thread()
        logger.debug(
            'Subscriber#%s relay shutdown done',
            self.relay_sub_id,
        )
        if do_unsubscribe:
            self.unsubscribe()

    def subscribe(self) -> None:
        assert self.relay_sub_id and self.relay_send
        self.event_queue.subscribe(self.relay_sub_id, self.relay_send)

    def unsubscribe(self) -> None:
        if self.relay_sub_id is None:
            logger.warning(
                'Relay called unsubscribe without an active subscription',
            )
            return
        try:
            self.event_queue.unsubscribe(self.relay_sub_id)
        except (BrokenPipeError, EOFError):
            pass
        finally:
            # self.relay_sub_id = None
            pass

    @staticmethod
    def relay(
            sub_id: str,
            shutdown: threading.Event,
            channel: connection.Connection,
            callback: Callable[[Dict[str, Any]], None],
    ) -> None:
        while not shutdown.is_set():
            try:
                if channel.poll(timeout=1):
                    ev = channel.recv()
                    if ev['event_name'] == eventNames.SUBSCRIBED:
                        logger.info(
                            'Subscriber#{0} subscribe ack received'.format(
                                sub_id,
                            ),
                        )
                    elif ev['event_name'] == eventNames.UNSUBSCRIBED:
                        logger.info(
                            'Subscriber#{0} unsubscribe ack received'.format(
                                sub_id,
                            ),
                        )
                        break
                    elif ev['event_name'] == eventNames.DISPATCHER_SHUTDOWN:
                        logger.info(
                            'Subscriber#{0} received dispatcher shutdown event'.format(
                                sub_id,
                            ),
                        )
                        break
                    else:
                        callback(ev)
            except queue.Empty:
                pass
            except EOFError:
                break
            except KeyboardInterrupt:
                break
        logger.debug('bbye!!!')

    def _start_relay_thread(self) -> None:
        self.relay_sub_id = uuid.uuid4().hex
        self.relay_shutdown = threading.Event()
        self.relay_recv, self.relay_send = multiprocessing.Pipe()
        self.relay_thread = threading.Thread(
            target=EventSubscriber.relay,
            args=(
                self.relay_sub_id, self.relay_shutdown,
                self.relay_recv, self.callback,
            ),
        )
        self.relay_thread.daemon = True
        self.relay_thread.start()

    def _stop_relay_thread(self) -> None:
        assert self.relay_thread and self.relay_shutdown and self.relay_recv and self.relay_send
        self.relay_shutdown.set()
        self.relay_thread.join()
        self.relay_recv.close()
        # Currently relay_send instance here in
        # subscriber is not the same as one received
        # by dispatcher.  This may cause file
        # descriptor leakage.  So we make a close
        # here explicit on our side of relay_send too.
        self.relay_send.close()
        self.relay_thread = None
        self.relay_shutdown = None
        self.relay_recv = None
        self.relay_send = None



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import sys
import socket
import logging
import argparse
from typing import TYPE_CHECKING, Any, Set, Optional, cast


try:
    if TYPE_CHECKING:  # pragma: no cover
        from ...common.types import HostPort
except ImportError:     # pragma: no cover
    pass

from .base import BaseSshTunnelHandler, BaseSshTunnelListener
from ...common.flag import flags


logger = logging.getLogger(__name__)


flags.add_argument(
    '--tunnel-hostname',
    type=str,
    default=None,
    help='Default: None. Remote hostname or IP address to which SSH tunnel will be established.',
)

flags.add_argument(
    '--tunnel-port',
    type=int,
    default=22,
    help='Default: 22. SSH port of the remote host.',
)

flags.add_argument(
    '--tunnel-username',
    type=str,
    default=None,
    help='Default: None. Username to use for establishing SSH tunnel.',
)

flags.add_argument(
    '--tunnel-ssh-key',
    type=str,
    default=None,
    help='Default: None. Private key path in pem format',
)

flags.add_argument(
    '--tunnel-ssh-key-passphrase',
    type=str,
    default=None,
    help='Default: None. Private key passphrase',
)

flags.add_argument(
    '--tunnel-remote-port',
    type=int,
    default=8899,
    help='Default: 8899. Remote port which will be forwarded locally for proxy.',
)


class SshTunnelListener(BaseSshTunnelListener):
    """Connects over SSH and forwards a remote port to local host.

    Incoming connections are delegated to provided callback."""

    def __init__(
        self,
        flags: argparse.Namespace,
        handler: BaseSshTunnelHandler,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        paramiko_logger = logging.getLogger('paramiko')
        paramiko_logger.setLevel(logging.WARNING)

        # pylint: disable=import-outside-toplevel
        from paramiko import SSHClient
        from paramiko.transport import Transport

        self.flags = flags
        self.handler = handler
        self.ssh: Optional[SSHClient] = None
        self.transport: Optional[Transport] = None
        self.forwarded: Set['HostPort'] = set()

    def start_port_forward(self, remote_addr: 'HostPort') -> None:
        assert self.transport is not None
        self.transport.request_port_forward(
            *remote_addr,
            handler=self.handler.on_connection,
        )
        self.forwarded.add(remote_addr)
        logger.debug('%s:%d forwarding successful...' % remote_addr)

    def stop_port_forward(self, remote_addr: 'HostPort') -> None:
        assert self.transport is not None
        self.transport.cancel_port_forward(*remote_addr)
        self.forwarded.remove(remote_addr)

    def setup(self) -> None:
        # pylint: disable=import-outside-toplevel
        from paramiko import SSHClient, AutoAddPolicy

        self.ssh = SSHClient()
        self.ssh.load_system_host_keys()
        self.ssh.set_missing_host_key_policy(AutoAddPolicy())
        self.ssh.connect(
            hostname=self.flags.tunnel_hostname,
            port=self.flags.tunnel_port,
            username=self.flags.tunnel_username,
            key_filename=self.flags.tunnel_ssh_key,
            passphrase=self.flags.tunnel_ssh_key_passphrase,
            compress=True,
            timeout=10,
            auth_timeout=7,
        )
        logger.debug(
            'SSH connection established to %s:%d...'
            % (
                self.flags.tunnel_hostname,
                self.flags.tunnel_port,
            ),
        )
        self.transport = self.ssh.get_transport()
        assert self.transport
        sock = cast(socket.socket, self.transport.sock)  # type: ignore[redundant-cast]
        # Enable TCP keep-alive
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
        # Keep-alive interval (in seconds)
        if sys.platform != 'darwin':
            sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, 30)
        # Keep-alive probe interval (in seconds)
        sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, 5)
        # Number of keep-alive probes before timeout
        sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPCNT, 5)
        self.start_port_forward(('', self.flags.tunnel_remote_port))

    def shutdown(self) -> None:
        for remote_addr in list(self.forwarded):
            self.stop_port_forward(remote_addr)
        self.forwarded.clear()
        if self.transport is not None:
            self.transport.close()
        if self.ssh is not None:
            self.ssh.close()
        self.handler.shutdown()

    def is_alive(self) -> bool:
        return self.transport.is_alive() if self.transport else False

    def is_active(self) -> bool:
        return self.transport.is_active() if self.transport else False



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import ssl
import logging
from typing import Optional

from .parser import HttpParser, httpParserTypes
from ..common.types import TcpOrTlsSocket
from ..common.utils import build_http_request, new_socket_connection
from ..common.constants import (
    HTTPS_PROTO, DEFAULT_TIMEOUT, DEFAULT_SSL_CONTEXT_OPTIONS,
)


logger = logging.getLogger(__name__)


def client(
    host: bytes,
    port: int,
    path: bytes,
    method: bytes,
    body: Optional[bytes] = None,
    conn_close: bool = True,
    scheme: bytes = HTTPS_PROTO,
    timeout: float = DEFAULT_TIMEOUT,
    content_type: bytes = b'application/x-www-form-urlencoded',
    verify: bool = True,
) -> Optional[HttpParser]:
    """HTTP Client"""
    request = build_http_request(
        method=method,
        url=path,
        headers={
            b'Host': host,
            b'Content-Type': content_type,
        },
        body=body,
        conn_close=conn_close,
    )
    try:
        conn = new_socket_connection((host.decode(), port))
    except Exception:
        # logger.exception('Cannot establish connection', exc_info=exc)
        return None
    sock: TcpOrTlsSocket = conn
    if scheme == HTTPS_PROTO:
        try:
            ctx = ssl.SSLContext(protocol=ssl.PROTOCOL_TLS_CLIENT)
            ctx.options |= DEFAULT_SSL_CONTEXT_OPTIONS
            ctx.check_hostname = verify
            ctx.verify_mode = ssl.CERT_NONE if not verify else ssl.CERT_REQUIRED
            ctx.load_default_certs()
            sock = ctx.wrap_socket(conn, server_hostname=host.decode())
        except Exception as exc:
            logger.exception('Unable to wrap', exc_info=exc)
            conn.close()
            return None
    parser = HttpParser(httpParserTypes.RESPONSE_PARSER)
    sock.settimeout(timeout)
    try:
        sock.sendall(request)
        while True:
            chunk = sock.recv(1024)
            if not chunk:
                break
            parser.parse(memoryview(chunk))
            if parser.is_complete:
                break
    finally:
        sock.close()
    return parser



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
from .proxy import Proxy, main, grout, sleep_loop, entry_point
from .testing import TestCase


__all__ = [
    # Grout entry point. See
    # https://jaxl.io/
    'grout',
    # PyPi package entry_point. See
    # https://github.com/abhinavsingh/proxy.py#from-command-line-when-installed-using-pip
    'entry_point',
    # Embed proxy.py. See
    # https://github.com/abhinavsingh/proxy.py#embed-proxypy
    'main',
    # Unit testing with proxy.py. See
    # https://github.com/abhinavsingh/proxy.py#unit-testing-with-proxypy
    'TestCase',
    'Proxy',
    # Utility exposed for demos
    'sleep_loop',
]



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
from typing import Tuple, Optional


class TlsServerKeyExchange:
    """TLS Server Key Exchange"""

    def __init__(self) -> None:
        self.data: Optional[bytes] = None

    def parse(self, raw: bytes) -> Tuple[bool, bytes]:
        self.data = raw
        return True, raw

    def build(self) -> bytes:
        assert self.data
        return self.data


class TlsClientKeyExchange:
    """TLS Client Key Exchange"""

    def __init__(self) -> None:
        self.data: Optional[bytes] = None

    def parse(self, raw: bytes) -> Tuple[bool, bytes]:
        return False, raw

    def build(self) -> bytes:
        assert self.data
        return self.data



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
from typing import Optional

from ..http import httpStatusCodes
from ..http.proxy import HttpProxyBasePlugin
from ..common.flag import flags
from ..http.parser import HttpParser
from ..common.utils import text_
from ..http.exception import HttpRequestRejected


flags.add_argument(
    '--filtered-upstream-hosts',
    type=str,
    default='facebook.com,www.facebook.com',
    help='Default: Blocks Facebook.  Comma separated list of IPv4 and IPv6 addresses.',
)


class FilterByUpstreamHostPlugin(HttpProxyBasePlugin):
    """Drop traffic by inspecting upstream host."""

    def before_upstream_connection(
            self, request: HttpParser,
    ) -> Optional[HttpParser]:
        if text_(request.host) in self.flags.filtered_upstream_hosts.split(','):
            raise HttpRequestRejected(
                status_code=httpStatusCodes.I_AM_A_TEAPOT,
                reason=b'I\'m a tea pot',
            )
        return request



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import json
from typing import TYPE_CHECKING, Any, Dict, List

from ..websocket import WebsocketFrame, WebSocketTransportBasePlugin
from ...core.event import EventSubscriber
from ...common.utils import bytes_


if TYPE_CHECKING:   # pragma: no cover
    from ..connection import HttpClientConnection


class InspectTrafficPlugin(WebSocketTransportBasePlugin):
    """Websocket API for inspect_traffic.ts frontend plugin."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.subscriber = EventSubscriber(
            self.event_queue,
            callback=lambda event: InspectTrafficPlugin.callback(
                self.client, event,
            ),
        )

    def methods(self) -> List[str]:
        return [
            'enable_inspection',
            'disable_inspection',
        ]

    def handle_message(self, message: Dict[str, Any]) -> None:
        if message['method'] == 'enable_inspection':
            # inspection can only be enabled if --enable-events is used
            if not self.flags.enable_events:
                self.reply({
                    'id': message['id'],
                    'response': 'not enabled',
                })
            else:
                self.subscriber.setup()
                self.reply(
                    {'id': message['id'], 'response': 'inspection_enabled'},
                )
        elif message['method'] == 'disable_inspection':
            self.subscriber.shutdown()
            self.reply({
                'id': message['id'],
                'response': 'inspection_disabled',
            })
        else:
            raise NotImplementedError()

    @staticmethod
    def callback(client: 'HttpClientConnection', event: Dict[str, Any]) -> None:
        event['push'] = 'inspect_traffic'
        client.queue(
            memoryview(
                WebsocketFrame.text(
                    bytes_(
                        json.dumps(event),
                    ),
                ),
            ),
        )



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
from typing import Any, Optional

from .client import SocksClientConnection
from ..core.base import BaseTcpServerHandler


class SocksProtocolHandler(BaseTcpServerHandler[SocksClientConnection]):
    """Reference https://www.openssh.com/txt/socks4.protocol"""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    @staticmethod
    def create(*args: Any) -> SocksClientConnection:
        return SocksClientConnection(*args)     # pragma: no cover

    def handle_data(self, data: memoryview) -> Optional[bool]:
        return super().handle_data(data)    # pragma: no cover



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.

    .. spelling::

       eventing
"""
import logging
import threading
import multiprocessing
from typing import Any, Optional

from .queue import EventQueue
from .dispatcher import EventDispatcher
from ...common.flag import flags
from ...common.constants import DEFAULT_ENABLE_EVENTS


logger = logging.getLogger(__name__)


flags.add_argument(
    '--enable-events',
    action='store_true',
    default=DEFAULT_ENABLE_EVENTS,
    help='Default: False.  Enables core to dispatch lifecycle events. '
    'Plugins can be used to subscribe for core events.',
)


class EventManager:
    """Event manager is a context manager which provides
    encapsulation around various setup and shutdown steps
    to start the eventing core.
    """

    def __init__(self) -> None:
        self.queue: Optional[EventQueue] = None
        self.dispatcher: Optional[EventDispatcher] = None
        self.dispatcher_thread: Optional[threading.Thread] = None
        self.dispatcher_shutdown: Optional[threading.Event] = None

    def __enter__(self) -> 'EventManager':
        self.setup()
        return self

    def __exit__(self, *args: Any) -> None:
        self.shutdown()

    def setup(self) -> None:
        self.queue = EventQueue(multiprocessing.Queue())
        self.dispatcher_shutdown = threading.Event()
        assert self.dispatcher_shutdown
        assert self.queue
        self.dispatcher = EventDispatcher(
            shutdown=self.dispatcher_shutdown,
            event_queue=self.queue,
        )
        self.dispatcher_thread = threading.Thread(
            target=self.dispatcher.run,
        )
        self.dispatcher_thread.start()
        logger.debug('Dispatcher#%d started', self.dispatcher_thread.ident)

    def shutdown(self) -> None:
        assert self.dispatcher_shutdown and self.dispatcher_thread
        self.dispatcher_shutdown.set()
        self.dispatcher_thread.join()
        logger.debug(
            'Dispatcher#%d shutdown',
            self.dispatcher_thread.ident,
        )



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.

    .. spelling::

       http
"""
from typing import TYPE_CHECKING, Any, Optional


if TYPE_CHECKING:   # pragma: no cover
    from ..parser import HttpParser


class HttpProtocolException(Exception):
    """Top level :exc:`HttpProtocolException` exception class.

    All exceptions raised during execution of HTTP request lifecycle MUST
    inherit :exc:`HttpProtocolException` base class. Implement
    ``response()`` method to optionally return custom response to client.
    """

    def __init__(self, message: Optional[str] = None, **kwargs: Any) -> None:
        super().__init__(message or 'Reason unknown')

    def response(self, request: 'HttpParser') -> Optional[memoryview]:
        return None  # pragma: no cover



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.

    .. spelling::

       dns
"""
import socket
from typing import Tuple, Optional

from ..http.proxy import HttpProxyBasePlugin
from ..common.types import HostPort


class CustomDnsResolverPlugin(HttpProxyBasePlugin):
    """This plugin demonstrate how to use your own custom DNS resolver."""

    def resolve_dns(self, host: str, port: int) -> Tuple[Optional[str], Optional[HostPort]]:
        """Here we are using in-built python resolver for demonstration.

        Ideally you would like to query your custom DNS server or even
        use :term:`DoH` to make real sense out of this plugin.

        The second parameter returned is None.  Return a 2-tuple to
        configure underlying interface to use for connection to the
        upstream server.
        """
        try:
            return socket.getaddrinfo(host, port, proto=socket.IPPROTO_TCP)[0][4][0], None
        except socket.gaierror:
            # Ideally we can also thrown HttpRequestRejected or HttpProtocolException here
            # Returning None simply fallback to core generated exceptions.
            return None, None



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import logging
from typing import Any, Optional

from .constants import DEFAULT_LOG_FILE, DEFAULT_LOG_LEVEL, DEFAULT_LOG_FORMAT


SINGLE_CHAR_TO_LEVEL = {
    'D': 'DEBUG',
    'I': 'INFO',
    'W': 'WARNING',
    'E': 'ERROR',
    'C': 'CRITICAL',
}


def single_char_to_level(char: str) -> Any:
    return getattr(logging, SINGLE_CHAR_TO_LEVEL[char.upper()[0]])


class Logger:
    """Common logging utilities and setup."""

    @staticmethod
    def setup(
            log_file: Optional[str] = DEFAULT_LOG_FILE,
            log_level: str = DEFAULT_LOG_LEVEL,
            log_format: str = DEFAULT_LOG_FORMAT,
    ) -> None:
        if log_file:    # pragma: no cover
            logging.basicConfig(
                filename=log_file,
                filemode='a',
                level=single_char_to_level(log_level),
                format=log_format,
            )
        else:
            logging.basicConfig(
                level=single_char_to_level(log_level),
                format=log_format,
            )



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import json
import logging
from typing import Any, Dict, List, Tuple

from .frame import WebsocketFrame
from .plugin import WebSocketTransportBasePlugin
from ..parser import HttpParser
from ..server import HttpWebServerBasePlugin, httpProtocolTypes
from ...common.utils import bytes_


logger = logging.getLogger(__name__)


class WebSocketTransport(HttpWebServerBasePlugin):
    """WebSocket transport framework."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.plugins: List[WebSocketTransportBasePlugin] = []
        # Registered methods and handler plugin
        self.methods: Dict[str, WebSocketTransportBasePlugin] = {}
        if b'WebSocketTransportBasePlugin' in self.flags.plugins:
            for klass in self.flags.plugins[b'WebSocketTransportBasePlugin']:
                p = klass(self.flags, self.client, self.event_queue)
                self.plugins.append(p)
                for method in p.methods():
                    self.methods[method] = p

    def routes(self) -> List[Tuple[int, str]]:
        return [
            (httpProtocolTypes.WEBSOCKET, r'/transport/$'),
        ]

    def handle_request(self, request: HttpParser) -> None:
        raise NotImplementedError()

    def on_websocket_open(self) -> None:
        for plugin in self.plugins:
            plugin.connected()

    def on_websocket_message(self, frame: WebsocketFrame) -> None:
        try:
            assert frame.data
            message = json.loads(frame.data)
        except UnicodeDecodeError:
            logger.error(frame.data)
            logger.info(frame.opcode)
            return

        method = message['method']
        if method == 'ping':
            self.reply({'id': message['id'], 'response': 'pong'})
        elif method in self.methods:
            self.methods[method].handle_message(message)
        else:
            logger.info(frame.data)
            logger.info(frame.opcode)
            self.reply({'id': message['id'], 'response': 'not_implemented'})

    def on_client_connection_close(self) -> None:
        for plugin in self.plugins:
            plugin.disconnected()

    def reply(self, data: Dict[str, Any]) -> None:
        self.client.queue(
            memoryview(
                WebsocketFrame.text(
                    bytes_(
                        json.dumps(data),
                    ),
                ),
            ),
        )



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.

    .. spelling::

       tcp
"""
import socket
import logging
import selectors
from abc import abstractmethod
from typing import Any, TypeVar, Optional

from ...core.work import Work
from ...common.flag import flags
from ...common.types import (
    Readables, Writables, TcpOrTlsSocket, SelectableEvents,
)
from ...common.utils import wrap_socket
from ...core.connection import TcpClientConnection
from ...common.constants import (
    DEFAULT_TIMEOUT, DEFAULT_KEY_FILE, DEFAULT_CERT_FILE,
    DEFAULT_MAX_SEND_SIZE, DEFAULT_CLIENT_RECVBUF_SIZE,
    DEFAULT_SERVER_RECVBUF_SIZE,
)


logger = logging.getLogger(__name__)


flags.add_argument(
    '--key-file',
    type=str,
    default=DEFAULT_KEY_FILE,
    help='Default: None. Server key file to enable end-to-end TLS encryption with clients. '
    'If used, must also pass --cert-file.',
)

flags.add_argument(
    '--cert-file',
    type=str,
    default=DEFAULT_CERT_FILE,
    help='Default: None. Server certificate to enable end-to-end TLS encryption with clients. '
    'If used, must also pass --key-file.',
)

flags.add_argument(
    '--client-recvbuf-size',
    type=int,
    default=DEFAULT_CLIENT_RECVBUF_SIZE,
    help='Default: ' + str(int(DEFAULT_CLIENT_RECVBUF_SIZE / 1024)) +
    ' KB. Maximum amount of data received from the '
    'client in a single recv() operation.',
)

flags.add_argument(
    '--server-recvbuf-size',
    type=int,
    default=DEFAULT_SERVER_RECVBUF_SIZE,
    help='Default: ' + str(int(DEFAULT_SERVER_RECVBUF_SIZE / 1024)) +
    ' KB. Maximum amount of data received from the '
    'server in a single recv() operation.',
)

flags.add_argument(
    '--max-sendbuf-size',
    type=int,
    default=DEFAULT_MAX_SEND_SIZE,
    help='Default: ' + str(int(DEFAULT_MAX_SEND_SIZE / 1024)) +
    ' KB. Maximum amount of data to flush in a single send() operation.',
)

flags.add_argument(
    '--timeout',
    type=int,
    default=DEFAULT_TIMEOUT,
    help='Default: ' + str(DEFAULT_TIMEOUT) +
    '.  Number of seconds after which '
    'an inactive connection must be dropped.  Inactivity is defined by no '
    'data sent or received by the client.',
)


T = TypeVar('T', bound=TcpClientConnection)


class BaseTcpServerHandler(Work[T]):
    """BaseTcpServerHandler implements Work interface.

    BaseTcpServerHandler lifecycle is controlled by Threadless core
    using asyncio.  If you want to also support threaded mode, also
    implement the optional run() method from Work class.

    An instance of BaseTcpServerHandler is created for each client
    connection.  BaseTcpServerHandler ensures that server is always
    ready to accept new data from the client.  It also ensures, client
    is ready to accept new data before flushing data to it.

    Most importantly, BaseTcpServerHandler ensures that pending buffers
    to the client are flushed before connection is closed.

    Implementations must provide::

       a. handle_data(data: memoryview) implementation
       b. Optionally, also implement other Work method
          e.g. initialize, is_inactive, shutdown
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.must_flush_before_shutdown = False
        logger.debug(
            'Work#%d accepted from %s',
            self.work.connection.fileno(),
            self.work.address,
        )

    def initialize(self) -> None:
        """Optionally upgrades connection to HTTPS,
        sets ``conn`` in non-blocking mode and initializes
        HTTP protocol plugins."""
        conn = self._optionally_wrap_socket(self.work.connection)
        conn.setblocking(False)
        logger.debug('Handling connection %s' % self.work.address)

    @abstractmethod
    def handle_data(self, data: memoryview) -> Optional[bool]:
        """Optionally return True to close client connection."""
        pass    # pragma: no cover

    async def get_events(self) -> SelectableEvents:
        events = {}
        # We always want to read from client
        # Register for EVENT_READ events
        if self.must_flush_before_shutdown is False:
            events[self.work.connection.fileno()] = selectors.EVENT_READ
        # If there is pending buffer for client
        # also register for EVENT_WRITE events
        if self.work.has_buffer():
            if self.work.connection.fileno() in events:
                events[self.work.connection.fileno()] |= selectors.EVENT_WRITE
            else:
                events[self.work.connection.fileno()] = selectors.EVENT_WRITE
        return events

    async def handle_events(
            self,
            readables: Readables,
            writables: Writables,
    ) -> bool:
        """Return True to shutdown work."""
        teardown = await self.handle_writables(
            writables,
        ) or await self.handle_readables(readables)
        if teardown:
            logger.debug(
                'Shutting down client {0} connection'.format(
                    self.work.address,
                ),
            )
        return teardown

    async def handle_writables(self, writables: Writables) -> bool:
        teardown = False
        if self.work.connection.fileno() in writables and self.work.has_buffer():
            logger.debug(
                'Flushing buffer to client {0}'.format(self.work.address),
            )
            self.work.flush(self.flags.max_sendbuf_size)
            if self.must_flush_before_shutdown is True and \
                    not self.work.has_buffer():
                teardown = True
                self.must_flush_before_shutdown = False
        return teardown

    async def handle_readables(self, readables: Readables) -> bool:
        teardown = False
        if self.work.connection.fileno() in readables:
            try:
                data = self.work.recv(self.flags.client_recvbuf_size)
            except ConnectionResetError:
                logger.info(
                    'Connection reset by client {0}'.format(
                        self.work.address,
                    ),
                )
                return True
            except TimeoutError:
                logger.info(
                    'Client recv timeout error {0}'.format(
                        self.work.address,
                    ),
                )
                return True
            if data is None:
                logger.debug(
                    'Connection closed by client {0}'.format(
                        self.work.address,
                    ),
                )
                teardown = True
            else:
                r = self.handle_data(data)
                if isinstance(r, bool) and r is True:
                    logger.debug(
                        'Implementation signaled shutdown for client {0}'.format(
                            self.work.address,
                        ),
                    )
                    if self.work.has_buffer():
                        logger.debug(
                            'Client {0} has pending buffer, will be flushed before shutting down'.format(
                                self.work.address,
                            ),
                        )
                        self.must_flush_before_shutdown = True
                    else:
                        teardown = True
        return teardown

    def _encryption_enabled(self) -> bool:
        return self.flags.keyfile is not None and \
            self.flags.certfile is not None

    def _optionally_wrap_socket(self, conn: socket.socket) -> TcpOrTlsSocket:
        """Attempts to wrap accepted client connection using provided certificates.

        Shutdown and closes client connection upon error.
        """
        if self._encryption_enabled():
            assert self.flags.keyfile and self.flags.certfile
            # TODO(abhinavsingh): Insecure TLS versions must not be accepted by default
            conn = wrap_socket(conn, self.flags.keyfile, self.flags.certfile)
            self.work._conn = conn
        return conn

    @staticmethod
    @abstractmethod
    def create(*args: Any) -> T:
        raise NotImplementedError()



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import os
import socket
import logging

from .base import BaseListener
from ...common.flag import flags


flags.add_argument(
    '--unix-socket-path',
    type=str,
    default=None,
    help='Default: None. Unix socket path to use.  ' +
    'When provided --host and --port flags are ignored',
)

logger = logging.getLogger(__name__)


class UnixSocketListener(BaseListener):
    """Unix socket domain listener."""

    def listen(self) -> socket.socket:
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.bind(self.flags.unix_socket_path)
        sock.listen(self.flags.backlog)
        sock.setblocking(False)
        logger.info(
            'Listening on %s' %
            self.flags.unix_socket_path,
        )
        return sock

    def shutdown(self) -> None:
        super().shutdown()
        if self.flags.unix_socket_path:
            os.remove(self.flags.unix_socket_path)



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import socket
import asyncio
import logging
from abc import abstractmethod
from typing import Any, TypeVar, Optional

from ...event import eventNames
from ..threadless import Threadless
from ....common.types import HostPort, TcpOrTlsSocket


T = TypeVar('T')

logger = logging.getLogger(__name__)


class ThreadlessFdExecutor(Threadless[T]):
    """A threadless executor which handles file descriptors
    and works with read/write events over a socket."""

    def work(self, *args: Any) -> None:
        fileno: int = args[0]
        addr: Optional[HostPort] = args[1]
        conn: Optional[TcpOrTlsSocket] = args[2]
        conn = conn or socket.socket(fileno=socket.dup(fileno))  # type: ignore[attr-defined]
        uid = '%s-%s-%s' % (self.iid, self._total, fileno)
        self.works[fileno] = self.create(uid, conn, addr)
        self.works[fileno].publish_event(
            event_name=eventNames.WORK_STARTED,
            event_payload={'fileno': fileno, 'addr': addr},
            publisher_id=self.__class__.__qualname__,
        )
        try:
            self.works[fileno].initialize()
            self._total += 1
        except Exception as e:
            logger.exception(   # pragma: no cover
                'Exception occurred during initialization',
                exc_info=e,
            )
            self._cleanup(fileno)

    @property
    @abstractmethod
    def loop(self) -> Optional[asyncio.AbstractEventLoop]:
        raise NotImplementedError()

    @abstractmethod
    def receive_from_work_queue(self) -> bool:
        raise NotImplementedError()

    @abstractmethod
    def work_queue_fileno(self) -> Optional[int]:
        raise NotImplementedError()



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import gzip
from typing import Any, Dict, Optional

from .codes import httpStatusCodes
from ..common.utils import build_http_response
from ..common.constants import (
    PROXY_AGENT_HEADER_KEY, PROXY_AGENT_HEADER_VALUE,
    DEFAULT_MIN_COMPRESSION_LENGTH,
)


PROXY_TUNNEL_ESTABLISHED_RESPONSE_PKT = memoryview(
    build_http_response(
        httpStatusCodes.OK,
        reason=b'Connection established',
        no_cl=True,
    ),
)

PROXY_TUNNEL_UNSUPPORTED_SCHEME = memoryview(
    build_http_response(
        httpStatusCodes.BAD_REQUEST,
        reason=b'Unsupported protocol scheme',
        conn_close=True,
        no_cl=True,
    ),
)

PROXY_AUTH_FAILED_RESPONSE_PKT = memoryview(
    build_http_response(
        httpStatusCodes.PROXY_AUTH_REQUIRED,
        reason=b'Proxy Authentication Required',
        headers={
            PROXY_AGENT_HEADER_KEY: PROXY_AGENT_HEADER_VALUE,
            b'Proxy-Authenticate': b'Basic',
        },
        body=b'Proxy Authentication Required',
        conn_close=True,
        no_cl=True,
    ),
)

BAD_REQUEST_RESPONSE_PKT = memoryview(
    build_http_response(
        httpStatusCodes.BAD_REQUEST,
        reason=b'BAD REQUEST',
        headers={
            b'Server': PROXY_AGENT_HEADER_VALUE,
        },
        conn_close=True,
    ),
)

NOT_FOUND_RESPONSE_PKT = memoryview(
    build_http_response(
        httpStatusCodes.NOT_FOUND,
        reason=b'NOT FOUND',
        headers={
            b'Server': PROXY_AGENT_HEADER_VALUE,
        },
        conn_close=True,
    ),
)

NOT_IMPLEMENTED_RESPONSE_PKT = memoryview(
    build_http_response(
        httpStatusCodes.NOT_IMPLEMENTED,
        reason=b'NOT IMPLEMENTED',
        headers={
            b'Server': PROXY_AGENT_HEADER_VALUE,
        },
        conn_close=True,
    ),
)

BAD_GATEWAY_RESPONSE_PKT = memoryview(
    build_http_response(
        httpStatusCodes.BAD_GATEWAY,
        reason=b'Bad Gateway',
        headers={
            PROXY_AGENT_HEADER_KEY: PROXY_AGENT_HEADER_VALUE,
        },
        body=b'Bad Gateway',
        conn_close=True,
        no_cl=True,
    ),
)


def okResponse(
        content: Optional[bytes] = None,
        headers: Optional[Dict[bytes, bytes]] = None,
        compress: bool = True,
        min_compression_length: int = DEFAULT_MIN_COMPRESSION_LENGTH,
        **kwargs: Any,
) -> memoryview:
    do_compress: bool = False
    if compress and content and len(content) > min_compression_length:
        do_compress = True
        if not headers:
            headers = {}
        headers.update({
            b'Content-Encoding': b'gzip',
        })
    return memoryview(
        build_http_response(
            200,
            reason=b'OK',
            headers=headers,
            body=gzip.compress(content)
            if do_compress and content
            else content,
            **kwargs,
        ),
    )


def permanentRedirectResponse(location: bytes) -> memoryview:
    return memoryview(
        build_http_response(
            httpStatusCodes.PERMANENT_REDIRECT,
            reason=b'Permanent Redirect',
            headers={
                b'Location': location,
                b'Content-Length': b'0',
            },
            conn_close=True,
        ),
    )


def seeOthersResponse(location: bytes) -> memoryview:
    return memoryview(
        build_http_response(
            httpStatusCodes.SEE_OTHER,
            reason=b'See Other',
            headers={
                b'Location': location,
                b'Content-Length': b'0',
            },
            conn_close=True,
        ),
    )



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.

    .. spelling::

       ip
"""
from typing import Optional

from ..http import httpStatusCodes
from ..http.proxy import HttpProxyBasePlugin
from ..common.flag import flags
from ..http.parser import HttpParser
from ..http.exception import HttpRequestRejected


flags.add_argument(
    '--filtered-client-ips-mode',
    type=str,
    default='blacklist',
    help='Default: blacklist.  Can be either "whitelist" (restrict access to specific IPs)'
    'or "blacklist" (allow everything except specific IPs).',
)

flags.add_argument(
    '--filtered-client-ips',
    type=str,
    default='127.0.0.1,::1',
    help='Default: 127.0.0.1,::1.  Comma separated list of IPv4 and IPv6 addresses.',
)


class FilterByClientIpPlugin(HttpProxyBasePlugin):
    """Allow only (whitelist) or Drop only (blacklist) traffic by inspecting incoming client IP address."""

    def before_upstream_connection(
            self, request: HttpParser,
    ) -> Optional[HttpParser]:
        assert not self.flags.unix_socket_path and self.client.addr
        assert self.flags.filtered_client_ips_mode in ('blacklist', 'whitelist')
        if self.flags.filtered_client_ips_mode == 'blacklist':
            if self.client.addr[0] in self.flags.filtered_client_ips.split(','):
                raise HttpRequestRejected(
                    status_code=httpStatusCodes.I_AM_A_TEAPOT,
                    reason=b'I\'m a tea pot',
                )
        elif self.flags.filtered_client_ips_mode == 'whitelist':
            if self.client.addr[0] not in self.flags.filtered_client_ips.split(','):
                raise HttpRequestRejected(
                    status_code=httpStatusCodes.I_AM_A_TEAPOT,
                    reason=b'I\'m a tea pot',
                )
        return request
